"""Utility functions need to run ceci"""

from contextlib import contextmanager
import sys
import os


def add_python_path(path, start):
    """add a path to the env var PYTHONPATH"""
    old = os.environ.get("PYTHONPATH", "")
    if start:
        new = path + ":" + old
    else:
        new = old + ":" + path
    os.environ["PYTHONPATH"] = new


def remove_python_path(path, start):
    """remove a path from PYTHONPATH"""
    p = os.environ.get("PYTHONPATH", "").split(":")
    if start:
        p.remove(path)
    else:
        remove_last(p, path)
    os.environ["PYTHONPATH"] = ":".join(p)


@contextmanager
def extra_paths(paths, start=True):
    """Add extra paths to PYTHONPATH while in a context"""

    # allow passing a single path or
    # a list of them
    if isinstance(paths, str):
        paths = paths.split()

    # On enter, add paths to both sys.path,
    # and the PYTHONPATH env var, so that subprocesses
    # can see it,
    # either the start or the end depending
    # on the start argument
    for path in paths:
        if start:
            sys.path.insert(0, path)
        else:
            sys.path.append(path)

        add_python_path(path, start)

    # Return control to caller
    try:
        yield
    # On exit, remove the paths
    finally:
        for path in paths:
            try:
                if start:
                    sys.path.remove(path)
                else:
                    remove_last(sys.path, path)
                # also remove env var entry
                remove_python_path(path, start)
            # If e.g. user has already done this
            # manually for some reason then just
            # skip
            except ValueError:
                pass


def remove_last(lst, item):
    """
    Removes (in-place) the last instance of item from the list lst.
    Raises ValueError if item is not in list

    Parameters
    ----------
    lst: List
        A list of anything
    item: object
        Item to be removed

    Returns
    -------
    None
    """
    tmp = lst[::-1]
    tmp.remove(item)
    lst[:] = tmp[::-1]


def embolden(text):
    """
    Make a string print in bold in a terminal using control codes.

    Note that if this text is read in other applications (not in a shell)
    then it probably won't work.


    Parameters
    ----------
    text: str
        Text to be emboldened

    Returns
    -------
    str
        Emboldened text
    """
    return "\033[1m" + text + "\033[0m"



"""Utility class to interface to workflow managers when using local resources, e.g., a laptop"""

from .site import Site
import socket
from ..minirunner import Node


class LocalSite(Site):
    """Object representing execution in the local environment, e.g. a laptop."""

    def command(self, cmd, sec):
        """Generate a complete command line to be run with the specified execution variables.

        This builds up the command from the core and adds any docker commands, env var settings,
        or mpirun calls.

        Parameters
        ----------
        cmd: str
            The core command to execute.

        sec: StageExecutionConfig
            sec objects contain info on numbers of processes, threads, etc, and container choices

        Returns
        -------
        full_cmd: str
            The complete decorated command to be executed.
        """

        mpi1 = f"{self.mpi_command} {sec.nprocess}" if sec.nprocess > 1 else ""
        mpi2 = "--mpi" if sec.nprocess > 1 else ""
        volume_flag = f"-v {sec.volume} " if sec.volume else ""
        paths = self.config.get("python_paths", [])

        # TODO: allow other container types here, like singularity
        if sec.image:
            # If we are setting python paths then we have to modify the executable
            # here.  This is because we need the path to be available right from the
            # start, in case the stage is defined in a module added by these paths.
            # The --env flags in docker/shifter overwrites an env var, and there
            # doesn't seem to be a way to just append to one, so we have to be a bit
            # roundabout to make this work, and invoke bash -c instead.
            paths_start = (
                "bash -c 'PYTHONPATH=$PYTHONPATH:" + (":".join(paths)) if paths else ""
            )
            paths_end = "'" if paths else ""
            return (
                f"docker run "
                f"--env OMP_NUM_THREADS={sec.threads_per_process} "
                f"{volume_flag} "
                f"--rm -it {sec.image} "
                f"{paths_start} "
                f"{mpi1} "
                f"{cmd} {mpi2} "
                f"{paths_end}"
            )
        else:
            # In the non-container case this is much easier
            paths_env = (
                "PYTHONPATH=" + (":".join(paths)) + ":$PYTHONPATH" if paths else ""
            )
            return (
                f"OMP_NUM_THREADS={sec.threads_per_process} "
                f"{paths_env} "
                f"{mpi1} "
                f"{cmd} {mpi2}"
            )

    def configure_for_parsl(self):
        """Utility function to set parsl configuration parameters"""
        from parsl.executors import ThreadPoolExecutor

        max_threads = self.config.get("max_threads", 4)
        executor = ThreadPoolExecutor(label="local", max_threads=max_threads)
        # executors = [executor]

        self.info["executor"] = executor

    def configure_for_mini(self):
        """Utility function to setup self for local execution"""
        import psutil

        # The default is to allow a single process
        # with as many cores as possible, but allow the
        # user to specify both max_processes and max_threads
        # to customize
        procs = self.config.get("max_processes", 1)
        cores_available = psutil.cpu_count(logical=False)
        threads_default = max(cores_available // procs, 1)
        cores = self.config.get("max_threads", threads_default)
        name = socket.gethostname()

        # Create a node, which actually just represents one process
        # here due to SLURM limitations
        nodes = [Node(f"{name}_{i}", cores) for i in range(procs)]
        self.info["nodes"] = nodes

    def configure_for_cwl(self):
        """Utility function to set CWL configuration parameters"""



from ceci.stage import PipelineStage
from ceci.config import StageParameter
from ceci_example.types import HDFFile
import numpy as np
from ceci.errors import *
import pytest
import h5py
import os

# TODO: test MPI facilities properly with:
# https://github.com/rmjarvis/TreeCorr/blob/releases/4.1/tests/mock_mpi.py


class MockCommunicator:
    def __init__(self, size, rank):
        self.size = size
        self.rank = rank

    def Get_size(self):
        return self.size

    def Get_rank(self):
        return self.rank


def test_construct():
    # This one should work
    class TestStage(PipelineStage):
        name = "test"
        inputs = [("inp1", HDFFile)]
        outputs = []
        config_options = {}

        def run(self):
            pass

    assert PipelineStage.get_stage("test") == TestStage

    # this fails if you execute this file directly as the qualified
    # name is different
    if __name__ != "__main__":
        assert TestStage.get_module().endswith("test_stage")

    s = TestStage({"config": "tests/config.yml", "inp1": "tests/test.hdf5"})

    assert s.rank == 0
    assert s.size == 1
    assert s.is_parallel() == False
    assert s.is_mpi() == False
    assert list(s.split_tasks_by_rank(["1", "2", "3"])) == ["1", "2", "3"]

    r = list(s.data_ranges_by_rank(1000, 100))
    assert r[0] == (0, 100)
    assert r[2] == (200, 300)

    r = list(s.iterate_hdf("inp1", "group1", ["x", "y", "z"], 10))
    for ri in r:
        s, e, ri = ri
        assert len(ri["x"] == 10)
    assert np.all(r[4][2]["z"] == [-80, -82, -84, -86, -88, -90, -92, -94, -96, -98])

    # Fake that we are processor 4/10
    comm = MockCommunicator(10, 4)
    s = TestStage({"config": "tests/config.yml", "inp1": "tests/test.hdf5"}, comm=comm)

    assert s.rank == 4
    assert s.size == 10
    assert s.is_parallel() == True
    assert s.is_mpi() == True
    assert list(s.split_tasks_by_rank("abcdefghijklmnopqrst")) == ["e", "o"]

    r = list(s.data_ranges_by_rank(10000, 100))
    assert r[0] == (400, 500)
    assert r[3] == (3400, 3500)

    r = list(s.iterate_hdf("inp1", "group1", ["x", "y", "z"], 10))
    assert len(r) == 1
    st, e, r = r[0]
    assert st == 40
    assert e == 50
    assert np.all(r["x"] == range(40, 50))

    r = list(s.iterate_hdf("inp1", "group1", ["x", "y", "z"], 10, parallel=False))
    for ri in r:
        st, e, ri = ri
        assert len(ri["x"]) == 10
    assert np.all(r[4][2]["z"] == [-80, -82, -84, -86, -88, -90, -92, -94, -96, -98])


    s = TestStage({"config": "tests/config.yml", "inp1": "tests/test.hdf5"})
    r = list(s.iterate_hdf("inp1", "group1", ["w", "x", "y", "z"], 10, longest=True))
    for i, ri in enumerate(r):
        st, e, ri = ri
        if i < 10:
            assert len(ri["x"]) == 10
            assert np.allclose(ri["x"], np.arange(10*i, 10*(i+1)))
            assert np.allclose(ri["w"], np.arange(10*i, 10*(i+1)))
        else:
            assert len(ri["x"]) == 0
            assert len(ri["w"]) == 10
            assert np.allclose(ri["w"], np.arange(10*i, 10*(i+1)))

    # check error message is raise appropriately if longest is not specified
    s = TestStage({"config": "tests/config.yml", "inp1": "tests/test.hdf5"})
    with pytest.raises(ValueError):
        r = list(s.iterate_hdf("inp1", "group1", ["w", "x", "y", "z"], 10, longest=False))


def test_make_stage():
    # This one should work
    class TestStage(PipelineStage):
        name = "test_copy"
        inputs = [("inp1", HDFFile)]
        outputs = [("out", HDFFile)]
        config_options = {"a":"b"}

        def run(self):
            pass

    stage = TestStage.make_stage(name="copy_of_test_stage", a='c', inp1='dummy', out="my_file.hdf5")
    assert stage.find_outputs('.')['out_copy_of_test_stage'] =='./my_file.hdf5'
    assert stage.config.a == 'c'


def test_parameter():
    # This one should work
    class TestStage(PipelineStage):
        name = "test_stage_param"
        inputs = [("inp1", HDFFile)]
        outputs = []
        config_options = dict(
            a=StageParameter(float, 5., msg="a float"),
            b=StageParameter(str, msg="a str"),
            c=StageParameter(list, [1,2,3], msg="a list"),
            d=StageParameter(list, [], msg="an empty list"),
        )

        def run(self):
            pass

    stage_1 = TestStage.make_stage(
        a=6., b='puffins are not extinct?', inp1='dummy',
        )
    assert stage_1.config.a == 6.
    assert stage_1.config.b == 'puffins are not extinct?'
    assert 1 in stage_1.config.c
    assert len(stage_1.config.d) == 0
    
    cmd = "TestStage", "--a", "6", "--b", "puffins are not extinct?", "--inp", "dummy"
    stage_1_cmd = TestStage(TestStage.parse_command_line(cmd))
    assert stage_1_cmd.config.a == 6.
    assert stage_1_cmd.config.b == 'puffins are not extinct?'


    # This one should not work
    class TestStage_2(PipelineStage):
        name = "test_stage_param_2"
        inputs = [("inp1", HDFFile)]
        outputs = []
        config_options = dict(a=StageParameter(float, None, msg="a float"))

        def run(self):
            pass

    with pytest.raises(ValueError):
        stage_2 = TestStage_2.make_stage(inp1='dummy')

    # This one should work
    class TestStage_3(PipelineStage):
        name = "test_stage_param_3"
        inputs = [("inp1", HDFFile)]
        outputs = []
        config_options = dict(a=StageParameter(float, None, required=False, msg="a float"))

        def run(self):
            pass

    stage_3 = TestStage_3.make_stage(inp1='dummy')
    assert stage_3.config.a is None

    par = StageParameter(int, 0, msg="a float")
    assert par.value == 0
    par.set_default(1)
    assert par.value == 1
    assert par.default == 1

    # This one should work
    class TestStage_4(PipelineStage):
        name = "test_stage_param_4"
        inputs = [("inp1", HDFFile)]
        outputs = []
        config_options = dict(a=stage_1.config)

        def run(self):
            pass

    stage_4 = TestStage_4.make_stage(inp1='dummy')
    assert stage_4.config.a == 5.


    stage_1.config.get('a').set_default(7.)
    # This one should work
    class TestStage_5(PipelineStage):
        name = "test_stage_param_5"
        inputs = [("inp1", HDFFile)]
        outputs = []
        config_options = dict(a=stage_1.config)

        def run(self):
            pass

    stage_5 = TestStage_4.make_stage(inp1='dummy')
    assert stage_5.config.a == 7.





def test_incomplete():
    class Alpha(PipelineStage):
        pass

    assert "Alpha" in PipelineStage.incomplete_pipeline_stages

    with pytest.raises(IncompleteStage):
        PipelineStage.get_stage("Alpha")


def test_auto_name():
    class Bravo(PipelineStage):
        inputs = []
        outputs = []
        config_options = {}

        def run(self):
            pass

    assert Bravo.name == "Bravo"
    assert PipelineStage.get_stage("Bravo") is Bravo


def test_duplicate():
    class Charlie(PipelineStage):
        inputs = []
        outputs = []

        def run(self):
            pass

    assert PipelineStage.get_stage("Charlie") is Charlie

    with pytest.raises(DuplicateStageName):

        class Charlie(PipelineStage):
            inputs = []
            outputs = []

            def run(self):
                pass

    with pytest.raises(DuplicateStageName):
        # Name it specified and duplicated
        class AlsoCharlie(PipelineStage):
            name = "Charlie"
            inputs = []
            outputs = []

            def run(self):
                pass

    # Should work okay as name is overwritten
    class Charlie(PipelineStage):
        name = "NotCharlie"
        inputs = []
        outputs = []

        def run(self):
            pass

    assert Charlie.name == "NotCharlie"
    assert PipelineStage.get_stage("NotCharlie") is Charlie


def test_explicit_config():
    with pytest.raises(ReservedNameError):

        class Delta(PipelineStage):
            inputs = [("config", None)]
            outputs = []
            config_options = {}

            def run(self):
                pass


def test_okay_abc_dupe_name():
    class Echo(PipelineStage):
        pass

    # okay, as the parent is intended as abstract
    class Echo(Echo):
        inputs = []
        outputs = []

        def run(self):
            pass

    assert Echo.name == "Echo"
    assert PipelineStage.get_stage("Echo") is Echo


def test_okay_abc_dupe_name2():
    class FoxtrotBase(PipelineStage):
        name = "Foxtrot"
        pass

    # okay, as the parent is intended as abstract
    class Foxtrot(FoxtrotBase):
        inputs = []
        outputs = []

        def run(self):
            pass

    assert Foxtrot.name == "Foxtrot"
    assert PipelineStage.get_stage("Foxtrot") is Foxtrot


def test_config_specified():
    # check for incomplete classes
    with pytest.raises(ReservedNameError):

        class Golf(PipelineStage):
            config = "golf"

    # check for complete classes
    with pytest.raises(ReservedNameError):

        class Golf(PipelineStage):
            config = "golf"
            inputs = []
            outputs = []

            def run(self):
                pass

def test_bool_flags():
    class Hotel(PipelineStage):
        inputs = []
        outputs = []
        config_options = {'xyz': bool}

    cmd = ["Hotel", "--config", "tests/config.yml"]

    # Basic case with a single flag
    h = Hotel(Hotel.parse_command_line(cmd + ["--xyz"]))
    assert h.config['xyz'] is True

    h = Hotel(Hotel.parse_command_line(cmd + ["--no-xyz"]))
    assert h.config['xyz'] is False

    # check latter takes precedence if both specified
    h = Hotel(Hotel.parse_command_line(cmd + ["--xyz", "--no-xyz"]))
    assert h.config['xyz'] is False

    h = Hotel(Hotel.parse_command_line(cmd + ["--no-xyz", "--xyz"]))
    assert h.config['xyz'] is True

    # flag is not optional here so must be set
    with pytest.raises(ValueError):
        h = Hotel(Hotel.parse_command_line(cmd))

    class Hotel2(PipelineStage):
        inputs = []
        outputs = []
        config_options = {'xyz': False}

    h = Hotel2(Hotel2.parse_command_line(cmd + ["--xyz"]))
    assert h.config['xyz'] is True

    h = Hotel2(Hotel2.parse_command_line(cmd + ["--no-xyz"]))
    assert h.config['xyz'] is False

    h = Hotel2(Hotel2.parse_command_line(cmd + ["--xyz", "--no-xyz"]))
    assert h.config['xyz'] is False

    h = Hotel2(Hotel2.parse_command_line(cmd + ["--no-xyz", "--xyz"]))
    assert h.config['xyz'] is True

    h = Hotel2(Hotel2.parse_command_line(cmd))
    assert h.config['xyz'] is False

    class Hotel3(PipelineStage):
        inputs = []
        outputs = []
        config_options = {'xyz': True}

    h = Hotel3(Hotel3.parse_command_line(cmd + ["--xyz"]))
    assert h.config['xyz'] is True

    h = Hotel3(Hotel3.parse_command_line(cmd + ["--no-xyz"]))
    assert h.config['xyz'] is False

    h = Hotel3(Hotel3.parse_command_line(cmd + ["--xyz", "--no-xyz"]))
    assert h.config['xyz'] is False

    h = Hotel3(Hotel3.parse_command_line(cmd + ["--no-xyz", "--xyz"]))
    assert h.config['xyz'] is True

    h = Hotel3(Hotel3.parse_command_line(cmd))
    assert h.config['xyz'] is True


def test_open_input():
    class India(PipelineStage):
        inputs = [("my_input", HDFFile)]
        outputs = []
        config_options = {}
    cmd = "India", "--config", "tests/config.yml", "--my_input", "tests/test.hdf5"

    # Testing without an alias
    ii = India(India.parse_command_line(cmd))
    assert os.path.exists(ii.get_input("my_input"))
    f = ii.open_input("my_input")
    print(f.keys())
    f.close()

    # Testing with an alias - config.yml defines an alias for my_input, my_alias
    ii = India.make_stage(name="IndiaCopy", my_input="tests/test.hdf5", config="tests/config.yml")

    print(ii.get_aliases())

    # These should work with or without the alias
    assert os.path.exists(ii.get_input("my_alias"))
    assert os.path.exists(ii.get_input("my_input"))

    # This works now
    f = ii.open_input("my_input")
    print(f.keys())
    f.close()

    f = ii.open_input("my_alias")
    print(f.keys())
    f.close()


def test_open_output():
    class Juliett(PipelineStage):
        inputs = []
        outputs = [("my_output", HDFFile)]
        config_options = {}
    cmd = "Juliett", "--config", "tests/config.yml", "--my_output", "tests/test_out.hdf5"

    # Testing without an alias
    jj = Juliett(Juliett.parse_command_line(cmd))
    #assert os.path.exists(jj.get_output("my_output"))
    f = jj.open_output("my_output")
    print(f.keys())
    f.close()

    # Testing with an alias
    jj = Juliett.make_stage(aliases=dict(my_output='my_alias'))

    print(jj.get_aliases())

    assert jj.get_output("my_output") == jj.get_output("my_alias")

    # This works now
    f = jj.open_output("my_output")
    print(f.keys())
    f.close()

    f = jj.open_output("my_alias")
    print(f.keys())
    f.close()

    # Testing with a new name
    jj = Juliett.make_stage(name="JuliettCopy")

    print(jj.get_aliases())

    assert jj.get_output("my_output") == jj.get_output("my_output_JuliettCopy")

    # Check we can open using the original name
    f = jj.open_output("my_output")
    print(f.keys())
    f.close()

    # Check with an alias specified for the output name
    f = jj.open_output("my_output_JuliettCopy")
    print(f.keys())
    f.close()


def core_test_map(comm):
    size = comm.size if comm else 1
    class Kilo(PipelineStage):
        name = f"Kilo_{size}"
        inputs = []
        outputs = []
        config_options = {"nproc_run": int}

        def run(self):
            self.power = 2
            self.my_count = 0
            x = [1, 2, 3, 4, 5, 6]
            y = self.map_tasks_by_rank(self.raise_to_power, x, allgather = True)
            assert y == [1, 4, 9, 16, 25, 36]
            assert self.my_count == len(x) // self.config["nproc_run"]

            self.power = 3
            self.my_count = 0
            y = self.map_tasks_by_rank(self.raise_to_power, x, allgather = False)

            if self.rank == 0:
                assert y == [1, 8, 27, 64, 125, 216]
            else:
                assert y is None
            assert self.my_count == len(x) // self.config["nproc_run"]

        def raise_to_power(self, x):
            self.my_count += 1
            return x ** self.power

    kk = Kilo.make_stage(comm=comm, nproc_run=size)
    kk.run()


def test_map():
    import mockmpi
    core_test_map(None)
    mockmpi.mock_mpiexec(2, core_test_map)
    mockmpi.mock_mpiexec(3, core_test_map)



def test_unknown_stage():
    with pytest.raises(StageNotFound):
        PipelineStage.get_stage("ThisStageIsDeliberatelyLeftBlank")


def test_wrong_mpi_flag():

    class LimaParallel(PipelineStage):
        name = f"LimaParallel"
        inputs = []
        outputs = []
        config_options = {}

        def run(self):
            pass

    class LimaSerial(LimaParallel):
        name = f"LimaSerial"
        parallel = False

    assert LimaParallel.parse_command_line(["LimaParallel", "--mpi"]).mpi
    assert not LimaParallel.parse_command_line(["LimaParallel"]).mpi

    with pytest.raises(ValueError):
        assert LimaSerial.parse_command_line(["LimaSerial", "--mpi"]).mpi


# could add more tests here for constructor, but the regression tests here and in TXPipe are
# pretty thorough.

if __name__ == "__main__":
    test_construct()
    test_wrong_mpi_flag()



import tempfile
import os
from functools import wraps


def in_temp_dir(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        with tempfile.TemporaryDirectory() as dirname:
            orig_dir = os.getcwd()
            os.chdir(dirname)
            try:
                res = f(*args, **kwargs)
            finally:
                os.chdir(orig_dir)
        return res

    return wrapper


@in_temp_dir
def get_temp_cwd():
    cwd = os.getcwd()
    print(cwd)
    return cwd


def test_in_temp_dir():
    d1 = os.getcwd()
    d2 = get_temp_cwd()
    d3 = os.getcwd()
    assert d2 != d1
    assert d3 == d1
    assert not os.path.exists(d2)



from parsl import clear
import tempfile
import os
import pytest
import subprocess

from ceci.pipeline import Pipeline
from ceci_example.example_stages import *
from ceci.config import StageConfig


def test_config():
    config_options = {'chunk_rows': 5000, 'something':float, 'free':None}
    config = StageConfig(**config_options)

    assert config['chunk_rows'] == 5000
    assert config.chunk_rows == 5000
    assert getattr(config, 'chunk_rows') == 5000

    config.chunk_rows = 133
    assert config.chunk_rows == 133

    config.free = 'dog'
    config.free = 42

    try:
        config.chunk_rows = 'a'
    except TypeError:
        pass
    else:
        raise RuntimeError("Failed to catch a type error")

    try:
        config['chunk_rows'] = 'a'
    except TypeError:
        pass
    else:
        raise RuntimeError("Failed to catch a type error")
    assert config.chunk_rows == 133

    config['new_par'] = 'abc'
    assert config['new_par'] == 'abc'
    assert config.get_type('new_par') == str

    config.reset()
    assert config.chunk_rows == 5000

    assert config.get_type('chunk_rows') == int

    values = config.values()
    for key, value in config.items():
        #assert value == config[key].value
        assert value in values

    def check_func(cfg, **kwargs):
        for k, v in kwargs.items():
            check_type = cfg.get_type(k)
            if k is not None and v is not None:
                assert check_type == type(v)

    check_func(config, **config)

    for k in iter(config):
        assert k in config

    
    

def test_interactive_pipeline():

    # Load the pipeline interactively, this is just a temp fix to
    # get the run_config and stage_config
    pipeline = Pipeline.read('tests/test.yml')
    # pipeline.run()

    dry_pipe = Pipeline.read('tests/test.yml', dry_run=True)
    dry_pipe.run()

    pipe2 = Pipeline.interactive()
    overall_inputs = {'DM':'./tests/inputs/dm.txt',
                      'fiducial_cosmology':'./tests/inputs/fiducial_cosmology.txt'}
    inputs = overall_inputs.copy()
    inputs['metacalibration'] = True
    inputs['config'] = None

    pipe2.pipeline_files.update(**inputs)
    pipe2.build_stage(PZEstimationPipe)
    pipe2.build_stage(shearMeasurementPipe, apply_flag=False)
    pipe2.build_stage(WLGCSelector, zbin_edges=[0.2, 0.3, 0.5], ra_range=[-5, 5])
    pipe2.build_stage(SysMapMaker)
    pipe2.build_stage(SourceSummarizer)
    pipe2.build_stage(WLGCCov, aliases=dict(covariance='covariance_copy'))
    pipe2.build_stage(WLGCRandoms)
    pipe2.build_stage(WLGCTwoPoint, name="WLGC2Point")
    pipe2.build_stage(WLGCSummaryStatistic, aliases=dict(covariance='covariance_copy'))

    assert len(pipe2.WLGCCov.outputs) == 1

    pipe2.initialize(overall_inputs, pipeline.run_config, pipeline.stages_config)

    pipe2.print_stages()
    pipe2.WLGCCov.print_io()

    assert pipe2['WLGCCov'] == pipe2.WLGCCov

    rpr = repr(pipe2.WLGCCov.config)

    path = pipe2.pipeline_files.get_path('covariance_copy')
    assert pipe2.pipeline_files.get_tag(path) == 'covariance_copy'
    assert pipe2.pipeline_files.get_type('covariance_copy') == pipe2.WLGCCov.get_output_type('covariance')

    pipe2.run()



def test_inter_pipe():

    pipe2 = Pipeline.interactive()
    overall_inputs = {'DM':'./tests/inputs/dm.txt',
                      'fiducial_cosmology':'./tests/inputs/fiducial_cosmology.txt'}
    inputs = overall_inputs.copy()
    inputs['config'] = None

    pipe2.pipeline_files.update(**inputs)

    pipe2.build_stage(PZEstimationPipe, name='bob')
    assert isinstance(pipe2.bob, PZEstimationPipe)
    pipe2.remove_stage('bob')

    assert not hasattr(pipe2, 'bob')







if __name__ == "__main__":
    test_config()
    test_interactive()



class DataFile:
    """
    A class representing a DataFile to be made by pipeline stages
    and passed on to subsequent ones.

    DataFile itself should not be instantiated - instead subclasses
    should be defined for different file types.

    These subclasses are used in the definition of pipeline stages
    to indicate what kind of file is expected.  The "suffix" attribute,
    which must be defined on subclasses, indicates the file suffix.

    The open method, which can optionally be overridden, is used by the
    machinery of the PipelineStage class to open an input our output
    named by a tag.

    """

    def __init__(self, path, mode, extra_provenance=None, validate=True, **kwargs):
        self.path = path
        self.mode = mode

        if mode not in ["r", "w"]:
            raise ValueError(f"File 'mode' argument must be 'r' or 'w' not '{mode}'")

        self.file = self.open(path, mode, **kwargs)

    @classmethod
    def open(cls, path, mode):
        """
        Open a data file.  The base implementation of this function just
        opens and returns a standard python file object.

        Subclasses can override to either open files using different openers
        (like fitsio.FITS), or, for more specific data types, return an
        instance of the class itself to use as an intermediary for the file.

        """
        return open(path, mode)

    @classmethod
    def make_name(cls, tag):
        if cls.suffix:
            return f"{tag}.{cls.suffix}"
        else:
            return tag


class HDFFile(DataFile):
    """
    A data file in the HDF5 format.
    Using these files requires the h5py package, which in turn
    requires an HDF5 library installation.

    """

    suffix = "hdf"
    format = "http://edamontology.org/format_3590"

    @classmethod
    def open(cls, path, mode, **kwargs):
        import warnings

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            import h5py
        return h5py.File(path, mode, **kwargs)


class FitsFile(DataFile):
    """
    A data file in the FITS format.
    Using these files requires the fitsio package.
    """

    suffix = "fits"
    format = "http://edamontology.org/format_2333"

    @classmethod
    def open(cls, path, mode, **kwargs):
        import fitsio

        # Fitsio doesn't have pure 'w' modes, just 'rw'.
        # Maybe we should check if the file already exists here?
        if mode == "w":
            mode = "rw"
        return fitsio.FITS(path, mode=mode, **kwargs)


class TextFile(DataFile):
    """
    A data file in plain text format.
    """

    suffix = "txt"
    format = "http://edamontology.org/format_2330"


class YamlFile(DataFile):
    """
    A data file in yaml format.
    """

    suffix = "yml"
    format = "http://edamontology.org/format_3750"



"""Module with core pipeline functionality """

import os
import sys
import collections
import yaml
import shutil
from abc import abstractmethod

from .stage import PipelineStage
from . import minirunner
from .sites import load, get_default_site
from .utils import embolden, extra_paths


def override_config(config, extra):
    """Override configuration parameters with extra parameters provided on command line

    Parameters
    ----------
    config : `Mapping`
        The original configuration, will be updated
    extra : `str`
        The additional arguments, in key=value pairs
    """
    print("Over-riding config parameters from command line:")

    for arg in extra:
        key, value = arg.split("=", 1)
        item = key.split(".")
        p = config
        print(f"    {key}: {value}")

        for x in item[:-1]:
            if x in p:
                p = p[x]
            else:
                p[x] = {}
                p = p[x]
        p[item[-1]] = yaml.safe_load(value)


class StageExecutionConfig:
    """
    This class stores information describing how an individual job is to be executed,
    for example how many cores it is run on, and where.  It does not store the
    job input or output information or anything like that.

    TODO: Consider attaching this object to the stage itself.

    Attributes
    ----------

    name: str
        The name of the stage
    class_name: str
        The name of the class of the stage
    module_name: str
        The name of the module for the stage
    site: Site object
        (default the global default) The site this stage is run on
    nprocess: int
        (default 1) The number of (usually MPI) processes to use for this task
    nodes: int
        (default 1) The number of nodes the processes should be spread over
    threads_per_process: int
        (default 1) The number of (usually OpenMP) threads to use per process.
    mem_per_process: float
        (defaut 2GB) The amount of memory in GB required for the job
    image: str
        (default is the site default) A docker image name to use for this task
    volume: str
        (default is the site default) Any volume mappings in the form
        /path/on/host:/path/on/container that the job needs
    stage_class : type
        (default is None) the class of the associated PipelineStage
    stage_obj : stage_class
        (default is None) and instance of stage_class, used to store configuration
    """

    def __init__(self, info):
        """Constructor, build from a dictionary used to set attributes

        Parameters
        ----------
        info : `Mapping'
            Dictionary used to initialize class, see class description
            for class attributes
        """
        # Core attributes - mandatory
        self.name = info["name"]
        self.class_name = info.get("classname", self.name)
        self.module_name = info.get("module_name")
        self.site = info.get("site", get_default_site())

        # Parallelism attributes - optional
        self.nprocess = info.get("nprocess", 1)
        self.nodes = info.get("nodes", 1)
        self.threads_per_process = info.get("threads_per_process", 1)  #
        self.mem_per_process = info.get("mem_per_process", 2)

        # Container attributes - optional.
        # There may be a default container for the entire site the
        # stage is run on, in which case use that if it is not overridden.
        self.image = info.get("image", self.site.config.get("image"))
        self.volume = info.get("volume", self.site.config.get("volume"))

        self.stage_class = None
        self.stage_obj = None

    @classmethod
    def create(cls, stage, **kwargs):
        """Construction method that builds a StageExecutionConfig
        from an existing PipelineStage object

        This is useful when building a Pipeline interactively

        Parameters
        ----------
        stage : `PipelineStage`
            The stage in question

        Keywords
        --------
        The keyword arguments are passed as a dictionary to the class constructor

        Returns
        -------
        sec : `StageExecutionConfig`
            The newly built object
        """
        info = kwargs.copy()
        info["name"] = stage.instance_name
        info["classname"] = stage.name
        info["module_name"] = stage.get_module()
        sec = cls(info)
        sec.set_stage_obj(stage)
        return sec

    def set_stage_obj(self, stage_obj):
        """Set the stage_obj attribute to a particular object

        Parameters
        ----------
        stage_obj : `PipelineClass`
            The object in question

        Raises
        ------
        TypeError : if stage_obj is not and instance of self.stage_class as
            determined by the self.class_name attribute
        """
        self.stage_class = PipelineStage.get_stage(self.class_name, self.module_name)
        if not isinstance(stage_obj, self.stage_class):  # pragma: no cover
            raise TypeError(f"{str(stage_obj)} is not a {str(self.stage_class)}")
        self.stage_obj = stage_obj

    def build_stage_class(self):
        """Set the stage_class attribute by finding
        self.class_name in the dictionary of classes from `Pipeline_stage`
        """
        self.stage_class = PipelineStage.get_stage(self.class_name, self.module_name)
        return self.stage_class

    def build_stage_object(self, args):
        """Build an instance of the PipelineStage by looking up the
        correct type and passing args to the constructor

        Parameters
        ----------
        args : `Mapping`
            Arguments passed to the constructor of self.stage_class

        Returns
        -------
        obj : `PipelineClass`
            The newly constructed object
        """
        if self.stage_class is None:  # pragma: no cover
            self.stage_class = PipelineStage.get_stage(
                self.class_name, self.module_name
            )
        self.stage_obj = self.stage_class(args)
        return self.stage_obj

    def generate_full_command(self, inputs, outputs, config):
        """Generate the full command needed to run this stage

        Parameters
        ----------
        inputs : `Mapping`
            Mapping of tags to paths for the stage inputs
        outputs : `Mapping`
            Mapping of tags to paths for the stage outputs
        config : `str`
            Path to file with stage configuration

        Returns
        -------
        command : str
            The command in question
        """
        if self.stage_obj is not None:
            aliases = self.stage_obj.get_aliases()
        else:
            aliases = None  # pragma: no cover
        if self.stage_class is None:
            self.build_stage_class()  # pragma: no cover
        core = self.stage_class.generate_command(
            inputs, config, outputs, aliases, self.name
        )
        return self.site.command(core, self)


class FileManager(dict):
    """Small class to manage files within a particular Pipeline

    This is a dict which is used for tag to path mapping,
    but has a couple of additional dicts to manage the reverse mapping and the
    tag to type mapping.


    The tag to path mapping is the thing that the Pipeline uses to set the
    input paths for downstream stages that use the outputs of earlier stages,
    i.e., everything in the pipeline can refer to a particular file by tag.

    The tag defaults to the input or output tag as define in the stage class attributes.
    However, in the case that we want multiple stages of the same class in a pipeline we
    have to alias the tags so that each stage can write to its own location (and so that
    downstream stages can pick up that location correctly)
    """

    def __init__(self):
        """Constructor, makes empty dictionaries"""
        self._tag_to_type = {}
        self._path_to_tag = {}
        dict.__init__(self)

    def __setitem__(self, tag, path):
        """Override dict.__setitem__() to also insert the reverse mapping"""
        dict.__setitem__(self, tag, path)
        self._path_to_tag[path] = tag

    def insert(self, tag, path=None, ftype=None):
        """Insert a file, including the path and the file type

        Parameters
        ----------
        tag : str
            The tag by which this file will be identified
        path : str
            The path to this file
        ftype : type
            The file type for this file
        """
        if path is not None:
            self[tag] = path
            self._path_to_tag[path] = tag
        if tag not in self._tag_to_type:
            self._tag_to_type[tag] = ftype

    def get_type(self, tag):
        """Return the file type associated to a given tag"""
        return self._tag_to_type[tag]

    def get_path(self, tag):
        """Return the path associated to a give tag"""
        return self[tag]

    def get_tag(self, path):
        """Return the tag associated to a given path"""
        return self._path_to_tag[path]

    def insert_paths(self, path_dict):
        """Insert a set of paths from a dict that has tag, path pairs"""
        for key, val in path_dict.items():
            self.insert(key, path=val)

    def insert_outputs(self, stage, outdir):
        """Insert a set of tags and associated paths and file types that are output by a stage"""
        stage_outputs = stage.find_outputs(outdir)
        for tag, ftype in stage.outputs:
            aliased_tag = stage.get_aliased_tag(tag)
            path = stage_outputs[aliased_tag]
            self.insert(aliased_tag, path=path, ftype=ftype)
        return stage_outputs


class Pipeline:
    """
    The Pipeline base class models the shared information and behaviour
    that pipelines need, no matter which workflow manager runs them.

    This includes constructing the pipeline in the first place from
    stages, checking that it is a valid pipeline, finding overall
    inputs and outputs for the pipeline, and constructing the command
    line for a given stage.

    Sub-classes run the pipeline using a given workflow manager.
    """

    def __init__(self, stages, launcher_config, **kwargs):
        """Construct a pipeline using configuraion information.

        Parameters
        ----------
        stages: list[dict]
            Information used to construct each stage and how it is run
        launcher_config: dict
            Any additional configuration that will be needed by the workflow
            management.  The base class does not use this.

        Keywords
        --------
        overall_inputs : Mapping or None
            The global inputs to the pipeline, as Mapping of tag:url
        modules : str
            Space seperated path of modules loaded for this pipeline
        """
        self.launcher_config = launcher_config

        self.overall_inputs = kwargs.get("overall_inputs", {}).copy()
        self.modules = kwargs.get("modules", "")

        # These are populated as we add stages below
        self.stage_execution_config = {}
        self.stage_names = []

        self.run_info = None
        self.run_config = None
        self.stages = None
        self.pipeline_files = FileManager()
        self.pipeline_outputs = None
        self.stages_config = None
        self.stage_config_data = None
        self.global_config = {}

        # Store the individual stage informaton
        for info in stages:
            self.add_stage(info)

    @staticmethod
    def create(pipe_config):
        """Create a Pipeline of a particular type, using the configuration provided

        Parameters
        ----------
        pipe_config : `Mapping`
            Dictionary of configuration parameters

        Returns
        -------
        pipeline : `Pipeline`
            The newly created pipeline
        """
        launcher_config = pipe_config.get("launcher")
        launcher_name = launcher_config["name"]
        stages_config = pipe_config["config"]
        stages = pipe_config["stages"]
        inputs = pipe_config["inputs"]
        modules = pipe_config["modules"]
        run_config = {
            "output_dir": pipe_config.get("output_dir", "."),
            "log_dir": pipe_config.get("log_dir", "."),
            "resume": pipe_config.get("resume", False),
            "flow_chart": pipe_config.get("flow_chart", ""),
        }

        launcher_dict = dict(cwl=CWLPipeline, parsl=ParslPipeline, mini=MiniPipeline)

        if pipe_config.get("flow_chart", False):
            pipeline_class = FlowChartPipeline
        elif pipe_config.get("dry_run", False):
            pipeline_class = DryRunPipeline
        else:
            try:
                pipeline_class = launcher_dict[launcher_name]
            except KeyError as msg:  # pragma: no cover
                raise KeyError(
                    f"Unknown pipeline launcher {launcher_name}, options are {list(launcher_dict.keys())}"
                ) from msg

        p = pipeline_class(
            stages, launcher_config, overall_inputs=inputs, modules=modules
        )
        p.initialize(inputs, run_config, stages_config)
        return p

    @staticmethod
    def interactive():
        """Build and return a pipeline specifically intended for interactive use"""
        launcher_config = dict(name="mini")
        return MiniPipeline([], launcher_config)

    @staticmethod
    def build_config(
        pipeline_config_filename, extra_config=None, dry_run=False, flow_chart=None
    ):
        """Build a configuration dictionary from a yaml file and extra optional parameters

        Parameters
        ----------
        pipeline_config_filename : str
            The path to the input yaml file
        extra_config : str
            A string with extra parameters in key=value pairs
        dry_run : bool
            A specfic flag to build a pipeline only from dry-runs

        Returns
        -------
        pipe_config : dict
            The resulting configuration
        """

        # YAML input file.
        # Load the text and then expand any environment variables
        with open(pipeline_config_filename) as config_file:
            raw_config_text = config_file.read()
        config_text = os.path.expandvars(raw_config_text)
        # Then parse with YAML
        pipe_config = yaml.safe_load(config_text)

        if extra_config:
            override_config(pipe_config, extra_config)

        # parsl execution/launcher configuration information
        launcher_config = pipe_config.setdefault("launcher", {"name": "mini"})

        # Launchers may need to know if this is a dry-run
        launcher_config["dry_run"] = dry_run
        pipe_config["dry_run"] = dry_run
        launcher_config["flow_chart"] = flow_chart
        pipe_config["flow_chart"] = flow_chart
        return pipe_config

    def __getitem__(self, name):
        """Get a particular stage by name"""
        try:
            return self.stage_execution_config[name].stage_obj
        except Exception as msg:  # pragma: no cover
            raise AttributeError(f"Pipeline does not have stage {name}") from msg

    def __getattr__(self, name):
        """Get a particular stage by name"""
        return self.__getitem__(name)

    def print_stages(self, stream=sys.stdout):
        """Print the list of stages in this pipeline to a stream"""
        for stage in self.stages:
            stream.write(f"{stage.instance_name:20}: {str(stage)}")
            stream.write("\n")

    @classmethod
    def read(cls, pipeline_config_filename, extra_config=None, dry_run=False):
        """Create a pipeline for a configuration dictionary from a yaml file and extra optional parameters

        Parameters
        ----------
        pipeline_config_filename : str
            The path to the input yaml file
        extra_config : str
            A string with extra parameters in key=value pairs
        dry_run : bool
            A specfic flag to build a pipeline only from dry-runs

        Returns
        -------
        pipeline : Pipeline
            The newly built pipeline
        """
        pipe_config = cls.build_config(pipeline_config_filename, extra_config, dry_run)
        paths = pipe_config.get("python_paths", [])
        if isinstance(paths, str):  # pragma: no cover
            paths = paths.split()

        modules = pipe_config["modules"].split()
        launcher_config = pipe_config.setdefault("launcher", {"name": "mini"})
        site_config = dict(name="local")
        site_config.update(**pipe_config.get("site"))
        # Pass the paths along to the site
        site_config["python_paths"] = paths
        load(launcher_config, [site_config])

        with extra_paths(paths):
            for module in modules:
                __import__(module)
        return cls.create(pipe_config)

    def add_stage(self, stage_info):
        """Add a stage to the pipeline.

        To begin with this stage is not connected to any others -
        that is determined later.

        There are two ways to add a stage.

        1) Passing in configuration information about the stage, see below

        2) Passing in an instance of a pipeline stage.  Typically this is done when
        building a pipeline interactively.

        The configuration info for this stage must contain at least
        the name of the stage and the name of the site where it is
        to be run.  It can also contain information for the
        StageExecutionConfig above describing parallelism
        and container usage.

        The stage_info can contain the following parameters:
            site: Site object
                The site this stage is run on
            nprocess: int
                (default 1) The number of (usually MPI) processes to use for this task
            nodes: int
                (default 1) The number of nodes the processes should be spread over
            threads_per_process: int
                (default 1) The number of (usually OpenMP) threads to use per process.
            mem_per_process: float
                (defaut 2GB) The amount of memory in GB required for the job
            image: str
                (default is the site default) A docker image name to use for this task
            volume: str
                (default is the site default) Any volume mappings in the form
                /path/on/host:/path/on/container that the job needs

        Parameters
        ----------
        stage_info: dict or PipelineStage
            Configuration information for this stage. See docstring for info.
        """
        if isinstance(stage_info, PipelineStage):
            sec = StageExecutionConfig.create(stage_info)
        else:
            sec = StageExecutionConfig(stage_info)
        self.stage_execution_config[sec.name] = sec
        self.stage_names.append(sec.name)
        if sec.stage_obj is None:
            return {}
        return self.pipeline_files.insert_outputs(sec.stage_obj, ".")

    def build_stage(self, stage_class, **kwargs):
        """Build a stage and add it to the pipeline

        Parameters
        ----------
        stage_class: type
            A subtype of `PipelineStage`, the class of the stage being build

        Returns
        -------
        stage_outputs: `dict`
            The names of the output files

        Notes
        -----
        The keyword arguments will be based to the `stage_class` constructor

        The output files produced by this stage will be added to the
        `Pipeline.pipeline_files` data member, so that they are available to later stages
        """
        kwcopy = kwargs.copy()
        kwcopy.update(**self.pipeline_files)
        stage = stage_class(kwcopy)
        return self.add_stage(stage)

    def remove_stage(self, name):
        """Delete a stage from the pipeline

        Parameters
        ----------
        name: str
            The name of the stage to remove.

        """
        self.stage_names.remove(name)
        del self.stage_execution_config[name]

    def get_stage_aliases(self, stage_name, stages_config=None):
        """Get the aliases for a particular stage

        Parameters
        ----------
        stage_name : str
            The name of the stage in question
        stages_config : Mapping or None
            Configurtion dictionary, used if stages have not been created yet

        Returns
        -------
        aliases : Mapping
            The aliases
        """
        stage = None
        sec = self.stage_execution_config.get(stage_name)
        if sec is not None:
            stage = sec.stage_obj
        if stage is not None:
            return stage.config.get("aliases", {})
        return stages_config.get(stage_name, {}).get("aliases", {})

    def ordered_stages(self, overall_inputs, stages_config=None):
        """Produce a linear ordering for the stages.

        Some stages within the pipeline might be runnable in parallel; this
        method does not analyze this, since different workflow managers will
        treat this differently.

        The stages in the pipeline are also checked for consistency, to avoid
        circular pipelines (A->B->C->A) and to ensure that all overall inputs
        needed in the pipeline are supplied from the overall inputs.

        The naive ordering algorithm used is faster when the stages are in
        the correct order to start with.  This won't matter unless you have
        a large number of stages.

        Parameters
        ----------
        overall_inputs: dict{str: str}
            Any inputs that do not need to be generated by the pipeline but are
            instead already supplied at the start.  Mapping is from tag -> path.

        Returns
        -------
        ordered_stages: list[PipelineStage]
            The pipeline stages in an order that can be run.

        """
        stage_names = self.stage_names[:]

        if stages_config:
            with open(stages_config) as fconfig:
                stage_config_data = yaml.safe_load(fconfig)
        else:
            stage_config_data = {}

        # First pass, get the classes for all the stages
        stage_classes = []
        for stage_name in stage_names:
            sec = self.stage_execution_config[stage_name]
            stage_classes.append(sec.build_stage_class())

        n = len(stage_names)

        # Check for a pipeline output that is already given as an input
        for stage_name in stage_names:
            stage_class = self.stage_execution_config[stage_name].stage_class
            stage_aliases = self.get_stage_aliases(stage_name, stage_config_data)
            for tag in stage_class.output_tags():
                aliased_tag = stage_aliases.get(tag, tag)
                if aliased_tag in overall_inputs:
                    raise ValueError(
                        f"Pipeline stage {stage_name} "
                        f"generates output {aliased_tag}, but "
                        "it is already an overall input"
                    )

        # Now check that the stage names are unique
        stage_set = set(stage_names)
        if len(stage_set) < len(stage_classes):
            raise ValueError("Some stages are included twice in your pipeline")

        # make a dict mapping each tag to the stages that need it
        # as an input. This is the equivalent of the adjacency matrix
        # in graph-speak
        dependencies = collections.defaultdict(list)
        for stage_name in stage_names:
            stage_class = self.stage_execution_config[stage_name].stage_class
            stage_aliases = self.get_stage_aliases(stage_name, stage_config_data)
            for tag in stage_class.input_tags():
                aliased_tag = stage_aliases.get(tag, tag)
                dependencies[aliased_tag].append(stage_name)

        # count the number of inputs required by each stage
        missing_input_counts = {}
        for stage_name in stage_names:
            stage_class = self.stage_execution_config[stage_name].stage_class
            missing_input_counts[stage_name] = len(stage_class.inputs)

        found_inputs = set()
        # record the stages which are receiving overall inputs
        for tag in overall_inputs:
            found_inputs.add(tag)
            for stage_name in dependencies[tag]:
                missing_input_counts[stage_name] -= 1

        # find all the stages that are ready because they have no missing inputs
        queue = [
            stage_name
            for stage_name in stage_names
            if missing_input_counts[stage_name] == 0
        ]
        ordered_stages = []

        all_inputs = overall_inputs.copy()

        # make the ordering
        while queue:
            # get the next stage that has no inputs missing
            stage_name = queue.pop()
            sec = self.stage_execution_config[stage_name]
            stage_class = sec.stage_class
            stage_config = stage_config_data.get(stage_name, {})
            stage_aliases = self.get_stage_aliases(stage_name, stage_config_data)
            stage_inputs = {}
            for tag in stage_class.input_tags():
                aliased_tag = stage_aliases.get(tag, tag)
                stage_inputs[aliased_tag] = all_inputs[aliased_tag]
            stage_config.update(stage_inputs)
            stage_config["config"] = stages_config
            if sec.stage_obj is None:
                stage = sec.build_stage_object(stage_config)
            else:
                stage = sec.stage_obj

            # for file that stage produces,
            stage_outputs = stage.find_outputs(".")
            for tag in stage.output_tags():
                # find all the next_stages that depend on that file
                aliased_tag = stage.get_aliased_tag(tag)
                found_inputs.add(aliased_tag)
                all_inputs[aliased_tag] = stage_outputs[aliased_tag]
                for next_stage in dependencies[aliased_tag]:
                    # record that the next stage now has one less
                    # missing dependency
                    missing_input_counts[next_stage] -= 1
                    # if that stage now has no missing stages
                    # then enqueue it
                    if missing_input_counts[next_stage] == 0:
                        queue.append(next_stage)
            ordered_stages.append(stage)

        # If any stages are still not in the list then there is a problem.
        # Try to diagnose it here.
        if len(ordered_stages) != n:
            stages_missing_inputs = [
                stage_name
                for (stage_name, count) in missing_input_counts.items()
                if count > 0
            ]
            msg1 = []
            for stage_name in stages_missing_inputs:
                stage_aliases = self.get_stage_aliases(stage_name, stage_config_data)
                stage_class = self.stage_execution_config[stage_name].stage_class
                missing_inputs = []
                for tag in stage_class.input_tags():
                    aliased_tag = stage_aliases.get(tag, tag)
                    if aliased_tag not in found_inputs:
                        missing_inputs.append(aliased_tag)
                missing_inputs = ", ".join(missing_inputs)
                msg1.append(f"Stage {stage_name} is missing input(s): {missing_inputs}")

            msg1 = "\n".join(msg1)
            raise ValueError(
                f"""
Some required inputs to the pipeline could not be found,
(or possibly your pipeline is cyclic):

{msg1}
"""
            )

        return ordered_stages

    def initialize(self, overall_inputs, run_config, stages_config):
        """Load the configuation for this pipeline

        Parameters
        ----------
        overall_inputs : `Mapping`
            A mapping from tag to path for all of the overall inputs needed by this pipeline
        run_config : `Mapping`
            Configuration parameters for how to run the pipeline
        stages_config: `str`
            File with stage configuration parameters

        Returns
        -------
        self.run_info : information on how to run the pipeline, as provided by sub-class `initiate_run` method
        self.run_config : copy of configuration parameters on how to run the pipeline
        """

        # Make a copy, since we maybe be modifying these
        self.overall_inputs = overall_inputs.copy()
        self.pipeline_files.insert_paths(self.overall_inputs)
        self.run_config = run_config.copy()

        self.stages_config = stages_config
        if self.stages_config is not None:
            with open(self.stages_config) as stage_config_file:
                self.stage_config_data = yaml.safe_load(stage_config_file)
        else:  # pragma: no cover
            self.stage_config_data = {}
        self.global_config = self.stage_config_data.pop("global", {})
        for v in self.stage_config_data.values():
            v.update(self.global_config)

        # Get the stages in the order we need.
        self.stages = self.ordered_stages(self.overall_inputs, self.stages_config)

        # Initiate the run.
        # This is an implementation detail for the different subclasses to store
        # necessary information about the run if necessary.
        # Usually, the arguments are ignored, but they are provided in case a class needs to
        # do something special with any of them.
        self.run_info = self.initiate_run(self.overall_inputs)

        # make sure output directories exist
        os.makedirs(run_config["output_dir"], exist_ok=True)
        os.makedirs(run_config["log_dir"], exist_ok=True)

        for stage in self.stages:
            # If we are in "resume" mode and the pipeline has already been run
            # then we re-use any existing outputs.  User is responsible for making
            # sure they are complete!

            if self.should_skip_stage(stage):
                stage.already_finished()
                self.pipeline_files.insert_outputs(stage, run_config["output_dir"])

            # Otherwise, run the pipeline and register any outputs from the
            # pipe element.
            else:
                stage_outputs = self.enqueue_job(stage, self.pipeline_files)
                self.pipeline_files.insert_paths(stage_outputs)

        return self.run_info, self.run_config

    def run(self):
        """Run the pipeline are return the execution status"""
        status = self.run_jobs()
        # When the
        self.pipeline_outputs = self.find_all_outputs()
        return status

    def find_all_outputs(self):
        """Find all the outputs associated to this pipeline

        Returns
        -------
        outputs : dict
            A dictionary of tag : path pairs will all of this Pipeline's outputs
        """
        outputs = {}
        for stage in self.stages:
            stage_outputs = stage.find_outputs(self.run_config["output_dir"])
            outputs.update(stage_outputs)
        return outputs

    @abstractmethod
    def initiate_run(self, overall_inputs):  # pragma: no cover
        """Setup the run and return any global information about how to run the pipeline"""
        raise NotImplementedError()

    @abstractmethod
    def enqueue_job(self, stage, pipeline_files):  # pragma: no cover
        """Setup the job for a single stage, and return stage specific information"""
        raise NotImplementedError()

    @abstractmethod
    def run_jobs(self):  # pragma: no cover
        """Actually run all the jobs and return the execution status"""
        raise NotImplementedError()

    def should_skip_stage(self, stage):
        """Return true if we should skip a stage because it is finished and we are in resume mode"""
        return stage.should_skip(self.run_config)

    def save(self, pipefile, stagefile=None, reduce_config=False):
        """Save this pipeline state to a yaml file

        Paramaeters
        -----------
        pipeline: str
            Path to the file were we save this
        stagefile: str
            Optional path to where we save the configuration file
        reduce_config: bool
            If true, reduce the configuration by parsing out the inputs, outputs and global params
        """
        pipe_dict = {}
        stage_dict = {}
        pipe_info_list = []
        if stagefile is None:
            stagefile = os.path.splitext(pipefile)[0] + "_config.yml"
        if self.run_config is not None:
            pipe_dict.update(**self.run_config)
        pipe_dict["config"] = stagefile
        if reduce_config:
            stage_dict["global"] = self.global_config
        site = None
        for key, val in self.stage_execution_config.items():
            if val.stage_obj is None:  # pragma: no cover
                raise ValueError(f"Stage {key} has not been built, can not save")
            if site is None:
                site = val.site.config
            pipe_stage_info = dict(
                name=val.name,
                classname=val.class_name,
                nprocess=val.nprocess,
                module_name=val.module_name,
            )
            if val.threads_per_process != 1:
                pipe_stage_info["threads_per_process"] = val.threads_per_process
            pipe_info_list.append(pipe_stage_info)
            stage_dict[key] = val.stage_obj.get_config_dict(
                self.global_config, reduce_config=reduce_config
            )

        module_list = list(
            {
                val[0].get_module().split(".")[0]
                for val in PipelineStage.pipeline_stages.values()
            }
        )
        not_first = False
        if self.modules:
            not_first = True
        for m_ in module_list:
            if not_first:
                self.modules += " "
            self.modules += f"{m_}"
            not_first = True
        pipe_dict["modules"] = self.modules
        pipe_dict["inputs"] = self.overall_inputs
        pipe_dict["stages"] = pipe_info_list
        pipe_dict["site"] = site
        with open(pipefile, "w") as outfile:
            try:
                yaml.dump(pipe_dict, outfile)
            except Exception as msg:  # pragma: no cover
                print(f"Failed to save {str(pipe_dict)} because {msg}")
        with open(stagefile, "w") as outfile:
            try:
                yaml.dump(stage_dict, outfile)
            except Exception as msg:  # pragma: no cover
                print(f"Failed to save {str(stage_dict)} because {msg}")

    def make_flow_chart(self, filename):
        import pygraphviz

        # Make a graph object
        graph = pygraphviz.AGraph(directed=True)

        # Nodes we have already added
        seen = set()

        # Add overall pipeline inputs
        for inp in self.overall_inputs.keys():
            graph.add_node(inp, shape="box", color="gold", style="filled")
            seen.add(inp)

        for stage in self.stages:
            # add the stage itself
            graph.add_node(
                stage.instance_name, shape="ellipse", color="orangered", style="filled"
            )
            # connect that stage to its inputs
            for inp, _ in stage.inputs:
                inp = stage.get_aliased_tag(inp)
                if inp not in seen:  # pragma: no cover
                    graph.add_node(inp, shape="box", color="skyblue", style="filled")
                    seen.add(inp)
                graph.add_edge(inp, stage.instance_name, color="black")
            # and to its outputs
            for out, _ in stage.outputs:
                out = stage.get_aliased_tag(out)
                if out not in seen:
                    graph.add_node(out, shape="box", color="skyblue", style="filled")
                    seen.add(out)
                graph.add_edge(stage.instance_name, out, color="black")

        # finally, output the stage to file
        if filename.endswith(".dot"):
            graph.write(filename)
        else:
            graph.draw(filename, prog="dot")


class DryRunPipeline(Pipeline):
    """A pipeline subclass which just does a dry-run, showing which commands
    would be executed.

    See the base class for almost all behaviour.

    No additional attributes.
    """

    def initiate_run(self, overall_inputs):
        return []

    def should_skip_stage(self, stage):
        return False

    def enqueue_job(self, stage, pipeline_files):
        outputs = stage.find_outputs(self.run_config["output_dir"])
        sec = self.stage_execution_config[stage.instance_name]

        cmd = sec.generate_full_command(pipeline_files, outputs, self.stages_config)

        # Replace the first instance of the stage name with bold
        # text, but only if we are printing to screen. This helps the
        # eye pick out the stage you want to run.
        if sys.stdout.isatty():
            cmd = cmd.replace(stage.instance_name, embolden(stage.instance_name), 1)

        self.run_info.append(cmd)
        return outputs

    def run_jobs(self):
        for cmd in self.run_info:
            print(cmd)
            print("\n")
        return 0

    def find_all_outputs(self):
        return {}


class FlowChartPipeline(DryRunPipeline):
    def run_jobs(self):
        filename = self.run_config["flow_chart"]
        self.make_flow_chart(filename)
        return 0


class ParslPipeline(Pipeline):
    """A subclass of Pipeline that uses Parsl to manage workflow."""

    def initiate_run(self, overall_inputs):
        return []  # list of futures

    def enqueue_job(self, stage, pipeline_files):

        from parsl.data_provider.files import File

        # log_dir = self.run_config["log_dir"]
        # convert the command into an app
        app = self.generate_app(stage, self.run_config)

        # Convert the dicts of inputs/outputs to the list that
        # parsl wants.
        # The inputs that exist already need to be converted into Parsl File objects.
        # The ones that don't stay as data futures
        inputs1 = stage.find_inputs(pipeline_files)
        inputs = [
            File(val) if isinstance(val, str) else val for val in inputs1.values()
        ]
        inputs.append(File(self.stages_config))
        # The outputs are just strings.  python dicts are now ordered,
        # so this works okay.
        outputs = [
            File(f) for f in stage.find_outputs(self.run_config["output_dir"]).values()
        ]

        # have parsl queue the app
        future = app(inputs=inputs, outputs=outputs)
        self.run_info.append((stage.instance_name, future))
        return {tag: future.outputs[i] for i, tag in enumerate(stage.output_tags())}

    def run_jobs(self):
        from parsl.app.errors import BashExitFailure

        log_dir = self.run_config["log_dir"]
        # Wait for the final results, from all files
        for stage_name, future in self.run_info:
            try:
                # This waits for b/g pipeline completion.
                future.result()
            # Parsl emits this on any non-zero status code.
            except BashExitFailure:
                stdout_file = f"{log_dir}/{stage_name}.err"
                stderr_file = f"{log_dir}/{stage_name}.out"
                sys.stderr.write(
                    f"""
*************************************************
Error running pipeline stage {stage_name}.

Standard output and error streams below.

*************************************************

Standard output:
----------------

"""
                )
                if os.path.exists(stdout_file):
                    with open(stdout_file) as _stdout:
                        sys.stderr.write(_stdout.read())
                else:  # pragma: no cover
                    sys.stderr.write("STDOUT MISSING!\n\n")

                sys.stderr.write(
                    """
*************************************************

Standard error:
----------------

"""
                )

                if os.path.exists(stderr_file):
                    with open(stderr_file) as _stderr:
                        sys.stderr.write(_stderr.read())
                else:  # pragma: no cover
                    sys.stderr.write("STDERR MISSING!\n\n")
                return 1
        return 0

    def generate_app(self, stage, run_config):
        """Build a parsl app that wraps this pipeline stage.

        This object is passed onto parsl for execution.

        Parameters
        ----------
        stage: PipelineStage
            The stage to wrap

        log_dir: str
            The directory to put the logs in.

        sec: StageExecutionConfig
            The execution info for this stage.

        Returns
        -------
        app: App
            A parsl app object

        """
        import parsl

        module = stage.get_module()
        module = module.split(".")[0]

        inputs = {}
        outputs = {}

        # Parsl wants a function with the inputs and outputs
        # extracted from lists.  That function should iself return
        # a string representing the bash cmd line.
        # We build up all these components here.

        # Parsl wants our functions to take their input/output paths
        # from inputs[0], inputs[1], etc.
        for i, inp in enumerate(stage.input_tags()):
            inputs[inp] = f"{{inputs[{i}]}}"
        for i, out in enumerate(stage.output_tags()):
            outputs[out] = f"{{outputs[{i}]}}"

        # The last input file is always the config file
        config_index = len(stage.input_tags())
        config = f"{{inputs[{config_index}]}}"

        # This includes all the "mpirun" stuff.
        sec = self.stage_execution_config[stage.instance_name]
        executor = sec.site.info["executor"]

        # Construct the command line call
        core = sec.generate_full_command(inputs, outputs, config)
        cmd1 = sec.site.command(core, sec)
        log_dir = run_config["log_dir"]

        # We will be exec'ing this here.  We can't just define it inline
        # because it just gets too painful with all the i/o names, so instead
        # we build and exec a string.
        template = f"""
@parsl.app.app.bash_app(executors=[executor])
def {stage.instance_name}(inputs, outputs, stdout='{log_dir}/{stage.instance_name}.out', stderr='{log_dir}/{stage.instance_name}.err'):
    cmd = '{cmd1}'.format(inputs=inputs, outputs=outputs)
    print("Launching command:")
    print(cmd, " 2> {log_dir}/{stage.instance_name}.err 1> {log_dir}/{stage.instance_name}.out")
    return cmd
"""
        print(template)

        # local variables for creating this function.
        d = {"executor": executor.label, "cmd1": cmd1}
        exec(template, {"parsl": parsl}, d)  # pylint: disable=exec-used

        # Return the function itself.
        return d[stage.instance_name]


class MiniPipeline(Pipeline):
    """A pipeline subclass that uses Minirunner, a sub-module
    of ceci, to run.

    Minirununer is a small tool I wrote suitable for interactive
    jobs on cori.  It launches jobs locally (not through a batch system),
    which parsl can also do, but it has a simple and clearer (to me at least)
    understanding of available nodes and cores.

    """

    def __init__(self, *args, **kwargs):
        """Create a MiniRunner Pipeline

        In addition to parent initialization parameters (see the
        Pipeline base class), this subclass can take these optional
        keywords.

        Parameters
        ----------
        callback: function(event_type: str, event_info: dict)
            A function called when jobs launch, complete, or fail,
            and when the pipeline aborts.  Can be used for tracing
            execution.  Default=None.

        sleep: function(t: float)
            A function to replace time.sleep called in the pipeline
            to wait until the next time to check process completion
            Most normal usage will not need this.  Default=None.
        """
        self.callback = kwargs.pop("callback", None)
        self.sleep = kwargs.pop("sleep", None)
        super().__init__(*args, **kwargs)

    def build_dag(self, jobs):
        """Build a directed acyclic graph of a set of stages.

        The DAG is represented by a list of jobs that each other job
        depends on.  If all a job's dependencies are complete
        then it can be run.

        Fun fact: the word "dag" is also Australian slang for a
        "dung-caked lock of wool around the hindquarters of a sheep".
        and is used as a semi-affectionate insult.

        Parameters
        ----------

        stages: list[PipelineStage]
            A list of stages to generate the DAG for

        """
        depend = {}

        # for each stage in our pipeline ...
        for stage in self.stages[:]:
            if stage.instance_name not in jobs:
                continue
            job = jobs[stage.instance_name]
            depend[job] = []
            # check for each of the inputs for that stage ...
            for tag in stage.input_tags():
                aliased_tag = stage.get_aliased_tag(tag)
                for potential_parent in self.stages[:]:
                    # if that stage is supplied by another pipeline stage
                    if potential_parent.instance_name not in jobs:  # pragma: no cover
                        continue
                    potential_parent_tags = [
                        potential_parent.get_aliased_tag(tag_)
                        for tag_ in potential_parent.output_tags()
                    ]
                    if aliased_tag in potential_parent_tags:
                        depend[job].append(jobs[potential_parent.instance_name])
        return depend

    def initiate_run(self, overall_inputs):
        jobs = {}
        stages = []
        return jobs, stages

    def enqueue_job(self, stage, pipeline_files):
        sec = self.stage_execution_config[stage.instance_name]
        outputs = stage.find_outputs(self.run_config["output_dir"])
        cmd = sec.generate_full_command(pipeline_files, outputs, self.stages_config)
        job = minirunner.Job(
            stage.instance_name,
            cmd,
            cores=sec.threads_per_process * sec.nprocess,
            nodes=sec.nodes,
        )
        self.run_info[0][stage.instance_name] = job
        self.run_info[1].append(stage)
        return outputs

    def run_jobs(self):
        jobs, _ = self.run_info
        # Now the jobs have all been queued, build them into a graph
        graph = self.build_dag(jobs)
        nodes = get_default_site().info["nodes"]
        log_dir = self.run_config["log_dir"]

        # This pipeline, being mainly for testing, can only
        # run at a single site, so we can assume here that the
        # sites are all the same
        sec = self.stage_execution_config[self.stage_names[0]]
        nodes = sec.site.info["nodes"]

        # Run under minirununer
        runner = minirunner.Runner(
            nodes, graph, log_dir, callback=self.callback, sleep=self.sleep
        )
        interval = self.launcher_config.get("interval", 3)
        try:
            runner.run(interval)
        except minirunner.FailedJob as error:
            sys.stderr.write(
                f"""
*************************************************
Error running pipeline stage {error.job_name}.

Standard output and error streams in {log_dir}/{error.job_name}.out
*************************************************
"""
            )
            return 1

        return 0


class CWLPipeline(Pipeline):
    """Export the pipeline as Common Workflow Language files and optionally run it
    with cwltool or another CWL-aware runner.

    """

    @staticmethod
    def make_inputs_file(
        stages, overall_inputs, stages_config, inputs_file
    ):  # pylint: disable=missing-function-docstring

        # find out the class of the file objects.  This is a bit ugly,
        # but the only way we record this now it in the stages.
        input_types = {}
        for stage in stages:
            for tag, ftype in stage.inputs:
                if (tag in overall_inputs) and (tag not in input_types):
                    input_types[tag] = ftype

        inputs = {}
        # First we build up the file inputs, each with this dictionary
        # information.
        for tag, filepath in overall_inputs.items():
            ftype = input_types[tag]
            filepath = os.path.abspath(filepath)
            d = {"class": "File", "path": filepath, "format": ftype.format}
            inputs[tag] = d

        # CWL also wants the config information passed through in an inputs
        # file, so it is all collected together.
        with open(stages_config) as _stages_config_file:
            stage_config_data = yaml.safe_load(_stages_config_file)
        global_config = stage_config_data.pop("global", {})

        # For each stage, we check if any of its config information
        # is set in the config file
        for stage in stages:
            # There might be nothing if no options are needed.
            this_stage_config = stage_config_data.get(stage.instance_name, {})
            # Record only keys that have been set.  If any are missing
            # it is an error that will be noticed later.
            for key in stage.config_options:
                val = this_stage_config.get(key, global_config.get(key))
                if val is not None:
                    inputs[f"{key}@{stage.instance_name}"] = val

        inputs["config"] = {
            "class": "File",
            "path": os.path.abspath(stages_config),
            # YAML file indicator:
            "format": "http://edamontology.org/format_3750",
        }

        # Save to the inputs file
        with open(inputs_file, "w") as f:
            yaml.dump(inputs, f)

    def initiate_run(self, overall_inputs):
        from cwl_utils.parser.cwl_v1_0 import Workflow

        wf = Workflow([], [], [], cwlVersion="v1.0")

        cwl_dir = self.launcher_config["dir"]
        os.makedirs(cwl_dir, exist_ok=True)

        # Write the inputs files
        inputs_file = f"{cwl_dir}/cwl_inputs.yml"
        self.make_inputs_file(
            self.stages, overall_inputs, self.stages_config, inputs_file
        )

        # CWL treats overall inputs differently, and requires
        # that we include the config file in there too
        self.overall_inputs["config"] = self.stages_config

        return {
            "workflow": wf,
            "cwl_dir": cwl_dir,
            "inputs_file": inputs_file,
            # keeps track of overall pipeline inputs we have already found
            "workflow_inputs": set(),
            "workflow_outputs": {},
        }

    def enqueue_job(self, stage, pipeline_files):
        from cwl_utils.parser.cwl_v1_0 import WorkflowStep, WorkflowStepInput
        from cwl_utils.parser.cwl_v1_0 import WorkflowOutputParameter, InputParameter

        cwl_dir = self.run_info["cwl_dir"]
        workflow = self.run_info["workflow"]
        log_dir = self.run_config["log_dir"]

        # Create a CWL representation of this step
        cwl_tool = stage.generate_cwl(log_dir)
        with open(f"{cwl_dir}/{stage.instance_name}.cwl", "w") as f:
            yaml.dump(cwl_tool.save(), f)

        # Load that representation again and add it to the pipeline
        step = WorkflowStep(stage.instance_name, [], [], run=f"{cwl_tool.id}.cwl")

        # For CWL these inputs are a mix of file and config inputs,
        # so not he same as the pipeline_files we usually see
        for inp in cwl_tool.inputs:

            if inp.id in self.overall_inputs:
                name = inp.id
            # If this input is an putput from an earlier stage
            # then it takes its name based on that
            elif inp.id in pipeline_files:
                name = pipeline_files[inp.id] + "/" + inp.id
            # otherwise if it's a config option we mangle
            # it to avod clashes
            elif inp.id in stage.config_options:
                name = f"{inp.id}@{cwl_tool.id}"
            # otherwise just leave it as-is
            else:  # pragma: no cover
                name = inp.id

            # If it's an overall input to the entire pipeline we
            # record that.  Or a configuration option.
            # And we don't want things that we've already recorded.
            if (
                (inp.id in self.overall_inputs) or (inp.id in stage.config_options)
            ) and (name not in self.run_info["workflow_inputs"]):
                self.run_info["workflow_inputs"].add(name)
                # These are the overall inputs to the enture pipeline.
                # Convert them to CWL input parameters
                cwl_inp = InputParameter(
                    name, label=inp.label, type=inp.type, format=inp.format
                )
                cwl_inp.default = inp.default

                # Bypassing cwlgen type check in case of arrays
                if isinstance(inp.type, dict):
                    cwl_inp.type = inp.type

                # record that these are overall pipeline inputs
                workflow.inputs.append(cwl_inp)

            # Record that thisis an input to the step.
            step.in_.append(WorkflowStepInput(id=inp.id, source=name))

        # Also record that we want all the pipeline outputs
        for tag, ftype in stage.outputs:
            # Record the expected output for this tag
            step.out.append(tag)

            # Also record that each file is an output to the entire pipeline
            cwl_out = WorkflowOutputParameter(
                tag,
                outputSource=f"{step.id}/{tag}",
                type="File",
                format=ftype.__name__,
            )
            workflow.outputs.append(cwl_out)

        # Also capture stdout and stderr as outputs
        cwl_out = WorkflowOutputParameter(
            f"{step.id}@stdout",
            outputSource=f"{step.id}/{step.id}@stdout",
            label="stdout",
            type="File",
        )
        step.out.append(f"{step.id}@stdout")
        workflow.outputs.append(cwl_out)

        cwl_out = WorkflowOutputParameter(
            f"{step.id}@stderr",
            outputSource=f"{step.id}/{step.id}@stderr",
            type="File",
        )
        step.out.append(f"{step.id}@stderr")
        workflow.outputs.append(cwl_out)

        # This step is now ready - add it to the list
        workflow.steps.append(step)

        # In CWL our data elemnts dict just records which step each
        # output is made in
        return {tag: step.id for tag in stage.output_tags()}

    def run_jobs(self):
        workflow = self.run_info["workflow"]
        cwl_dir = self.run_info["cwl_dir"]
        output_dir = self.run_config["output_dir"]
        log_dir = self.run_config["log_dir"]
        inputs_file = self.run_info["inputs_file"]

        with open(f"{cwl_dir}/pipeline.cwl", "w") as f:
            yaml.dump(workflow.save(), f)

        # If 'launcher' is defined, use that
        launcher = self.launcher_config.get(
            "launch",
            f"cwltool --outdir {output_dir} " "--preserve-environment PYTHONPATH",
        )
        if launcher == "cwltool":  # pragma: no cover
            launcher = (
                f"cwltool --outdir {output_dir} " "--preserve-environment PYTHONPATH"
            )

        if launcher:
            # need to include the CWD on the path for CWL as it
            # runs in an isolated directory
            pypath = os.environ.get("PYTHONPATH", "")
            os.environ["PYTHONPATH"] = pypath + ":" + os.getcwd()
            cmd = f"{launcher} {cwl_dir}/pipeline.cwl {inputs_file}"
            print(cmd)
            status = os.system(cmd)
            if pypath:  # pragma: no cover
                os.environ["PYTHONPATH"] = pypath
            else:
                del os.environ["PYTHONPATH"]

        # Parsl insists on putting everything in the same output directory,
        # both logs and file outputs.
        # we need to move those

        if status == 0:
            for step in self.run_info["workflow"].steps:
                shutil.move(f"{output_dir}/{step.id}.out", f"{log_dir}/{step.id}.out")
                shutil.move(f"{output_dir}/{step.id}.err", f"{log_dir}/{step.id}.err")

        return status



"""Utility class to interface to workflow managers with site-specific configuration"""

import warnings


class Site:
    """Object representing execution at a specific site"""

    default_mpi_command = "mpirun -n"

    def __init__(self, config):
        """Constructor, takes a dict of configuration parameters"""
        self.mpi_command = config.get("mpi_command", self.default_mpi_command)
        self.info = {}
        self.config = config

    def check_import(self, launcher):  # pylint: disable=no-self-use
        """Make sure that required libraries can be imported, and raise ImportError if they can not

        Parameters
        ----------
        launcher : str
            The launcher being built

        Raises
        ------
        ImportError : If the libraries for the requested launcher can not be built
        """
        requirements = {
            "parsl": ["parsl"],
            "cwl": ["cwl_utils", "cwltool"],
            "mini": ["psutil"],
        }
        if launcher not in requirements:  # pragma: no cover
            raise ValueError(f"Unknown launcher '{launcher}'")
        missing = []
        libs = requirements[launcher]
        for lib in libs:
            try:
                with warnings.catch_warnings():
                    warnings.filterwarnings("ignore", category=DeprecationWarning)
                    __import__(lib)
            except ImportError:  # pragma: no cover
                missing.append(lib)
        if missing:  # pragma: no cover
            missing = ", ".join(missing)
            raise ImportError(
                f"You must install these libraries "
                f"to use the {launcher} launcher: {missing}"
            )

    def configure_for_launcher(self, launcher):
        """Check to see if the given launcher is supported at the site in questions

        Notes
        -----
        This looks for a method called configure_for_{launcher} and will raise a ValueError
        if the method associated to the requested launcher does not exist.
        """

        self.check_import(launcher)
        configure = getattr(self, f"configure_for_{launcher}", None)
        if configure is None:  # pragma: no cover
            raise ValueError(
                f"Site {self} does not know how to configure for launcher {launcher}"
            )
        configure()



""" A small module with functionality to handle configuration in a way that works
for both interactive and ballistic applications """


def cast_value(dtype, value):  # pylint: disable=too-many-return-statements
    """Casts an input value to a particular type

    Parameters
    ----------
    dtype : type
        The type we are casting to
    value : ...
        The value being cast

    Returns
    -------
    out : ...
        The object cast to dtype

    Raises
    ------
    TypeError if neither value nor dtype are None and the casting fails

    Notes
    -----
    This will proceed in the following order
        1.  If dtype is None it will simply return value
        2.  If value is None it will return None
        3.  If value is an instance of dtype it will return value
        4.  If value is a Mapping it will use it as a keyword dictionary to the constructor of dtype, i.e., return dtype(**value)
        5.  It will try to pass value to the constructor of dtype, i.e., return dtype(value)
        6.  If all of these fail, it will raise a TypeError
    """
    # dtype is None means all values are legal
    if dtype is None:
        return value
    # value is None is always allowed
    if value is None:
        return None
    # if value is an instance of self.dtype, then return it
    if isinstance(value, dtype):
        return value
    # try the constructor of dtype
    try:
        return dtype(value)
    except (TypeError, ValueError):
        pass

    msg = f"Value of type {type(value)}, when {str(dtype)} was expected."
    raise TypeError(msg)


def cast_to_streamable(value):
    """Cast a value to something that yaml can easily stream"""
    if isinstance(value, StageParameter):
        return value.value
    return value


class StageParameter:
    """A small class to manage a single parameter with basic type checking"""

    def __init__(
        self, dtype=None, default=None, fmt="%s", required=None, msg="A parameter"
    ):
        """Build from keywords

        Parameters
        ----------
        dtype : `type` or `None`
            The data type for this parameter
        default : `dtype` or `None`
            The default value
        fmt : `str`
            A formatting string for printout and representation
        required : `bool` or `None`

        msg : `str`
            A help or docstring
        """
        self._help = msg
        self._format = fmt
        self._dtype = dtype
        self._default = default
        if required is not None:
            self._required = required
        else:
            self._required = self._default is None
        self._value = cast_value(self._dtype, self._default)

    @property
    def value(self):
        """Return the value"""
        return self._value

    @property
    def dtype(self):
        """Return the data type"""
        return self._dtype

    @property
    def default(self):
        """Return the default value"""
        return self._default

    @property
    def required(self):
        """Return the required flag"""
        return self._required

    def copy(self):
        """Return a copy of self"""
        return StageParameter(
            dtype=self._dtype,
            default=self._default,
            fmt=self._format,
            required=self._required,
            msg=self._help,
        )

    def set(self, value):
        """Set the value, raising a TypeError if the value is the wrong type"""
        self._value = cast_value(self._dtype, value)
        return self._value

    def set_to_default(self):
        """Set the value to the default"""
        self._value = cast_value(self._dtype, self._default)
        return self._value

    def set_default(self, default):
        """Set the default value"""
        self._default = default
        self.set_to_default()


class StageConfig(dict):
    """A small class to manage a dictionary of configuration parameters with basic type checking"""

    def __init__(self, **kwargs):
        """Build from keywords

        Note
        ----
        The keywords are used as keys for the configuration parameters

        The values are used to define the allowed data type and default values

        For each key-value pair:
        If the value is a type then it will define the data type and the default will be `None`
        If the value is a value then it will set the default value define the data type as type(value)
        """
        dict.__init__(self)
        for key, val in kwargs.items():
            param = None
            dtype = None
            default = None
            if val is None:
                pass
            elif isinstance(val, StageParameter):
                param = val.copy()
            elif isinstance(val, StageConfig):
                param = val.get(key).copy()
            elif isinstance(val, type):
                dtype = val
            else:
                dtype = type(val)
                default = val
            if param is None:
                param = StageParameter(dtype=dtype, default=default)
            self[key] = param

    def __str__(self):
        """Override __str__ casting to deal with `StageParameter` object in the map"""
        s = "{"
        for key, attr in dict.items(self):
            if isinstance(attr, StageParameter):
                val = attr.value
            else:
                val = attr
            s += f"{key}:{val},"
        s += "}"
        return s

    def __repr__(self):
        """A custom representation"""
        s = "StageConfig"
        s += self.__str__()
        return s

    def to_dict(self):
        """Forcibly return a dict where the values have been cast from StageParameter"""
        return {key: cast_to_streamable(value) for key, value in dict.items(self)}

    def __iter__(self):
        """Override the __iter__ to work with `StageParameter`"""
        d = self.to_dict()
        return iter(d)

    def __getitem__(self, key):
        """Override the __getitem__ to work with `StageParameter`"""
        attr = dict.__getitem__(self, key)
        if isinstance(attr, StageParameter):
            return attr.value
        return attr

    def __setitem__(self, key, value):
        """Override the __setitem__ to work with `StageParameter`"""
        if key in self:
            attr = dict.__getitem__(self, key)
            if isinstance(attr, StageParameter):
                return attr.set(value)
        dict.__setitem__(self, key, value)
        return value

    def __getattr__(self, key):
        """Allow attribute-like parameter access"""
        return self.__getitem__(key)

    def __setattr__(self, key, value):
        """Allow attribute-like parameter setting"""
        return self.__setitem__(key, value)

    def items(self):
        """Override items() to get the parameters values instead of the objects"""
        return [(key, cast_to_streamable(value)) for key, value in dict.items(self)]

    def values(self):
        """Override values() to get the parameters values instead of the objects"""
        return [cast_to_streamable(value) for value in dict.values(self)]

    def set_config(self, input_config, args):
        """Utility function to load configuration

        Parameters
        ----------
        input_config : `dict, (str, value)`
            `dict` with key-value pairs for all the parameters
        args : `dict, (str, value)`
            `dict` with key-value pairs for all the parameters that can serve as overrides
        """
        for key in self.keys():
            val = None
            if key in input_config:
                val = input_config[key]
            if args.get(key) is not None:
                val = args[key]
            if val is None:
                attr = self.get(key)
                if attr.required:
                    raise ValueError(f"Missing configuration option {key}")
                val = attr.default
            self.__setattr__(key, val)

        for key, val in input_config.items():
            if key in self:
                continue
            self[key] = val

        for key, val in args.items():
            if key in self:
                continue
            self[key] = val

    def reset(self):
        """Reset values to their defaults"""
        for _, val in dict.items(self):
            if isinstance(val, StageParameter):
                val.set_to_default()

    def get_type(self, key):
        """Get the type associated to a particular configuration parameter"""
        attr = dict.__getitem__(self, key)
        if isinstance(attr, StageParameter):
            return attr.dtype
        return type(attr)




from collections import OrderedDict

from ceci.sites.nersc import parse_int_set
from ceci.pipeline import override_config
from ceci.utils import embolden
from ceci.config import cast_value, cast_to_streamable


def test_parse_ints():
    assert parse_int_set("1,2,3") == set([1, 2, 3])
    assert parse_int_set("10-12") == set([10, 11, 12])
    assert parse_int_set("10-12,15,19-21") == set([10, 11, 12, 15, 19, 20, 21])


def test_override_config():
    config = {
        "a": "b",
        "c": {"d": "e"},
        "h": True,
        "i": 8,
        "j": 17.5,
    }
    override_config(config, ["a=a", "c.d=e", "f.x.y.z=g", "h=False", "i=9", "j=19.5"])

    assert config["a"] == "a"
    assert config["c"] == {"d": "e"}
    assert config["f"] == {"x": {"y": {"z": "g"}}}
    assert config["h"] is False
    assert config["i"] == 9
    assert config["j"] == 19.5

def test_embolden():
    x = 'hj6_9xx0'
    y = embolden(x)
    assert x in embolden(x)
    assert y[4:-4] == x


def test_cast_value():
    # dtype is None should allow any value
    assert cast_value(None, 5) == 5
    assert cast_value(None, "dog") == "dog"

    # value is None is always allowed
    assert cast_value(float, None) is None
    assert cast_value(str, None) is None

    # if isinstance(value, dtype) return value
    assert cast_value(float, 5.) == 5.
    assert cast_value(str, "dog") == "dog"

    # if isinstance(value, Mapping) return dtype(**value)
    odict = cast_value(dict, dict(key1='dog', key2=5))
    assert odict['key1'] == 'dog'
    assert odict['key2'] == 5

    # if dtype(value) works return that
    assert cast_value(float, 5) == 5.

    # catch errors
    try:
        cast_value(float, "dog")
    except TypeError:
        pass
    else:
        raise TypeError("Failed to catch type error")
    
    try:
        cast_value(int, [])
    except TypeError:
        pass
    else:
        raise TypeError("Failed to catch type error")
    

def test_cast_streamable():
    assert cast_to_streamable(dict(key='dog'))['key'] == 'dog'
    assert cast_to_streamable(OrderedDict([('key', 'dog')]))['key'] == 'dog'



"""Functions to run ceci from the command line"""

import os
import sys
import argparse
import subprocess
from .pipeline import Pipeline
from .sites import load, set_default_site, get_default_site
from .utils import extra_paths
import contextlib

# Add the current dir to the path - often very useful
sys.path.append(os.getcwd())

parser = argparse.ArgumentParser(
    description="Run a Ceci pipeline from a configuration file"
)
parser.add_argument(
    "pipeline_config", help="Pipeline configuration file in YAML format."
)
parser.add_argument(
    "--dry-run",
    action="store_true",
    help="Just print out the commands the pipeline would run without executing them",
)
parser.add_argument(
    "--flow-chart",
    type=str,
    default="",
    help="Make a flow chart image instead of running anything",
)
parser.add_argument(
    "extra_config",
    nargs="*",
    help="Over-ride the main pipeline yaml file e.g. launcher.name=cwl",
)


def run_pipeline(pipe_config):
    """Run a pipeline as defined by a particular configuration

    Parameters
    ----------
    pipe_config : `dict`
        The configuration dictionary

    Returns
    -------
    status : `int`
        Usual unix convention of 0 -> success, non-zero is an error code
    """
    with prepare_for_pipeline(pipe_config):
        p = Pipeline.create(pipe_config)
        status = p.run()
    return status


@contextlib.contextmanager
def prepare_for_pipeline(pipe_config):
    """
    Prepare the paths and launcher needed to read and run a pipeline.
    """

    # Later we will add these paths to sys.path for running here,
    # but we will also need to pass them to the sites below so that
    # they can be added within any containers or other launchers
    # that we use
    paths = pipe_config.get("python_paths", [])
    if isinstance(paths, str):  # pragma: no cover
        paths = paths.split()

    # Get information (maybe the default) on the launcher we may be using
    launcher_config = pipe_config.setdefault("launcher", {"name": "mini"})
    site_config = pipe_config.get("site", {"name": "local"})

    # Pass the paths along to the site
    site_config["python_paths"] = paths
    load(launcher_config, [site_config])

    # Python modules in which to search for pipeline stages
    modules = pipe_config.get("modules", "").split()

    # This helps with testing
    default_site = get_default_site()

    # temporarily add the paths to sys.path,
    # but remove them at the end
    with extra_paths(paths):

        # Import modules. We have to do this because the definitions
        # of the stages can be inside.
        for module in modules:
            __import__(module)

        try:
            yield
        finally:
            set_default_site(default_site)


def main():  # pragma: no cover
    """Main function called when ceci is invoked on the command line"""
    args = parser.parse_args()
    # If we are making a flow chart then we also set dry_run to stop
    # pre-scripts and post-scripts running
    if args.flow_chart:
        args.dry_run = True
    pipe_config = Pipeline.build_config(
        args.pipeline_config, args.extra_config, args.dry_run, args.flow_chart
    )
    status = run_pipeline(pipe_config)

    if status == 0:
        print("Pipeline successful.  Joy is sparked.")
    else:
        print("Pipeline failed.  No joy sparked.")
    return status


if __name__ == "__main__":  # pragma: no cover
    sys.exit(main())



import tempfile
import os
import pytest
import subprocess

from ceci.pipeline import Pipeline


def test_snapshot():

    # Read the pipeline
    pipeline = Pipeline.read('tests/test.yml')

    # Save it
    pipeline.save('test_save.yml')

    # Re-read it
    pipe_read = Pipeline.read('test_save.yml')

    # Introspect
    pipe_read.print_stages()
    pipe_read.WLGCCov.print_io()

    # And run
    pipe_read.run()
        
    pipeline.save('test_save2.yml', reduce_config=True)
    
    # Re-read it
    pipe_read2 = Pipeline.read('test_save2.yml')

    # Introspect
    pipe_read2.print_stages()
    pipe_read2.WLGCCov.print_io()



    
if __name__ == "__main__":
    test_snapshot()




import pytest
import subprocess
import os

def test_executable():
    subprocess.check_call(["ceci", "tests/test.yml"])

def test_dry_run():
    subprocess.check_call(["ceci", "--dry-run", "tests/test.yml",])

def test_flow_chart_run():
    subprocess.check_call(["ceci", "--flow-chart", "test-flow.png", "tests/test.yml",])
    assert os.path.exists("test-flow.png")

def test_profiling_and_memmon_flags():
    cmd = "python3 -m ceci_example PZEstimationPipe   --DM=./tests/inputs/dm.txt   --fiducial_cosmology=./tests/inputs/fiducial_cosmology.txt   --config=./tests/config.yml   --photoz_pdfs=./tests/outputs/photoz_pdfs.txt --memmon=1 --cprofile=profile.stats"
    subprocess.check_call(cmd.split())

def test_misuse_mpi():
    cmd = "python3 -m ceci_example PZEstimationPipe --mpi"
    with pytest.raises(subprocess.CalledProcessError):
        subprocess.check_call(cmd.split())


from ceci.stage import PipelineStage
import mockmpi

def core_dask(comm):
    class DaskTestStage(PipelineStage):
        name = "dasktest"
        dask_parallel = True
        inputs = []
        outputs = []
        config_options = {}

        def run(self):
            import dask.array as da
            arr = da.arange(100)
            x = arr.sum()
            x = x.compute()
            assert x == 4950


    args = DaskTestStage.parse_command_line(["dasktest", "--config", "tests/config.yml"])
    DaskTestStage.execute(args, comm=comm)

    # check that all procs get here
    if comm is not None:
        comm.Barrier()


def test_dask():
    core_dask(None)
    mockmpi.mock_mpiexec(3, core_dask)
    mockmpi.mock_mpiexec(5, core_dask)


if __name__ == '__main__':
    test_dask()



from ceci.main import run_pipeline
from parsl import clear
import tempfile
import os
import pytest
import subprocess

from ceci.pipeline import Pipeline


def run1(*config_changes, config_yaml="tests/test.yml", dry_run=False, expect_fail=False, expect_outputs=True, flow_chart=None):
    try:
        with tempfile.TemporaryDirectory() as dirname:
            out_dir = os.path.join(dirname, "output")
            log_dir = os.path.join(dirname, "logs")
            config = [f"output_dir={out_dir}", f"log_dir={log_dir}"]
            config += config_changes
            pipe_config = Pipeline.build_config(config_yaml, config, dry_run, flow_chart)
            status = run_pipeline(pipe_config)
            if expect_fail:
                assert status != 0
            else:
                assert status == 0
            if expect_outputs:
                assert os.path.exists(os.path.join(out_dir, "wlgc_summary_data.txt"))
                assert os.path.exists(os.path.join(log_dir, "WLGCSummaryStatistic.out"))

    finally:
        clear()


def test_run_mini():
    run1()


def test_run_dry_run():
    run1(dry_run=True, expect_fail=False, expect_outputs=False)

def test_flow_chart():
    run1(flow_chart="test.png", expect_outputs=False)

def test_flow_chart_dot():
    run1(flow_chart="test.dot", expect_outputs=False)


def test_run_parsl():
    run1("launcher.name=parsl", "launcher.max_threads=3")


def test_run_cwl():
    run1("launcher.name=cwl", "launcher.dir=tests/cwl") == 0


def test_run_namespace():
    run1(config_yaml="tests/test_namespace.yml", expect_outputs=False) == 0
  



if __name__ == "__main__":
    test_run_dry_run()
    test_run_parsl()
    test_run_mini()
    test_run_cwl()
    test_pre_script()



"""Module with core functionality for a single pipeline stage """

import pathlib
import os
import sys
from textwrap import dedent
import shutil
import cProfile
import pdb
import datetime

from abc import abstractmethod
from . import errors
from .monitor import MemoryMonitor
from .config import StageParameter, StageConfig, cast_to_streamable

SERIAL = "serial"
MPI_PARALLEL = "mpi"
DASK_PARALLEL = "dask"

IN_PROGRESS_PREFIX = "inprogress_"


class PipelineStage:
    """A PipelineStage implements a single calculation step within a wider pipeline.

    Each different type of analysis stage is represented by a subclass of this
    base class.  The base class handles the connection between different pipeline
    stages, and the execution of the stages within a workflow system (parsl),
    potentially in parallel (MPI).

    An instance of one of these classes represents an actual run of the stage,
    with the required inputs, outputs, and configuration specified.

    See documentation pages for more details.

    """

    parallel = True
    dask_parallel = False
    config_options = {}
    doc = ""
    allow_reload = False

    def __init__(self, args, comm=None):
        """Construct a pipeline stage, specifying the inputs, outputs, and configuration for it.

        The constructor needs a dict or namespace. It should include:
        - input paths (required)
        - config path (required)
        - output paths (optional but usual)
        - additional configuration (required if not specified elsewhere)

        Input and output paths should map tags to paths.
        Tags are strings, and the first elements in each item in the subclass's
        "inputs" and "output" attributes.
        e.g. for a subclass with:
            inputs = [('eggs', TextFile)]
            outputs = [('spam', TextFile)]
        the args could contain:
            {'eggs': 'inputs/eggs.txt',
             'spam': 'outputs/spam.txt' }
        If spam is not specified it will default to "./spam.txt"

        }

        The config should map "config" to a path where a YAML config file
        is located, e.g. {'config':'/path/to/config.yml'}

        Any config variables that are specified in the class's config attribute
        will be searched for first in args, then in the config file, and then
        by looking at any default value they have been given.
        If they have no default value (and just a type, like int, is listed), then
        it's an error if they are not specified somewhere.

        The execute method can instantiate and run the class together, with added bonuses
        like profiling and debugging tools.

        Parameters
        ----------
        args: dict or namespace
            Specification of input and output paths and any missing config options
        comm: MPI communicator
            (default is None) An MPI comm object to use in preference to COMM_WORLD
        """
        self._configs = StageConfig(**self.config_options)
        self._inputs = None
        self._outputs = None
        self._parallel = SERIAL
        self._comm = None
        self._size = 1
        self._rank = 0
        self._io_checked = False
        self.dask_client = None

        self.load_configs(args)
        if comm is not None:
            self.setup_mpi(comm)

    @classmethod
    def make_stage(cls, **kwargs):
        """Make a stage of a particular type"""
        kwcopy = kwargs.copy()
        kwcopy.setdefault("config", None)
        comm = kwcopy.pop("comm", None)
        name = kwcopy.get("name", None)
        aliases = kwcopy.pop("aliases", {})
        for input_ in cls.inputs:
            kwcopy.setdefault(input_[0], "None")
        if name is not None:
            for output_ in cls.outputs:  # pylint: disable=no-member
                outtag = output_[0]
                aliases[outtag] = f"{outtag}_{name}"
        kwcopy["aliases"] = aliases
        return cls(kwcopy, comm=comm)

    def get_aliases(self):
        """Returns the dictionary of aliases used to remap inputs and outputs
        in the case that we want to have multiple instance of this class in the pipeline"""
        return self.config.get("aliases", None)

    def get_aliased_tag(self, tag):
        """Returns the possibly remapped value for an input or output tag

        Parameter
        ---------
        tag : `str`
            The input or output tag we are checking

        Returns
        -------
        aliased_tag : `str`
            The aliases version of the tag
        """
        aliases = self.get_aliases()
        if aliases is None:
            return tag
        return aliases.get(tag, tag)

    @abstractmethod
    def run(self):  # pragma: no cover
        """Run the stage and return the execution status"""
        raise NotImplementedError("run")

    def load_configs(self, args):
        """
        Load the configuraiton

        Parameters
        ----------
        args: dict or namespace
            Specification of input and output paths and any missing config options
        """
        if not isinstance(args, dict):
            args = vars(args)

        # We alwys assume the config arg exists, whether it is in input_tags or not
        if "config" not in args:  # pragma: no cover
            raise ValueError("The argument --config was missing on the command line.")

        _name = args.get("name")
        if _name is not None:
            self._configs.name = _name

        # First, we extract configuration information from a combination of
        # command line arguments and optional 'config' file
        self._inputs = dict(config=args["config"])
        try:
            self.read_config(args)
        except Exception as error:
            error_class = type(error)
            msg = str(error)
            raise error_class(f"Error configuring {self.instance_name}: {msg}")
        self.check_io(args)

    def check_io(self, args=None):
        """
        Check the inputs and outputs.
        This function is seperate so that when Stages are configured interactively after
        construction then can invove this

        Parameters
        ----------
        args: dict or namespace
            Specification of input and output paths and any missing config options
        """

        # We first check for missing input files, that's a show stopper
        if self._io_checked:  # pragma: no cover
            return
        if args is None:  # pragma: no cover
            args = self.config

        missing_inputs = []
        for x in self.input_tags():
            val = args.get(x)
            aliased_tag = self.get_aliased_tag(x)

            if val is None:
                val = args.get(aliased_tag)

            if val is None:  # pragma: no cover
                missing_inputs.append(f"--{x}")
            else:
                self._inputs[aliased_tag] = val
        if missing_inputs:  # pragma: no cover
            missing_inputs = "  ".join(missing_inputs)
            raise ValueError(
                f"""

{self.instance_name} Missing these names on the command line:
    Input names: {missing_inputs}"""
            )

        # We prefer to receive explicit filenames for the outputs but will
        # tolerate missing output filenames and will default to tag name in
        # current folder (this is for CWL compliance)
        self._outputs = {}
        for i, x in enumerate(self.output_tags()):
            aliased_tag = self.get_aliased_tag(x)
            if args.get(x) is None:
                ftype = self.outputs[i][1]  # pylint: disable=no-member
                self._outputs[aliased_tag] = ftype.make_name(aliased_tag)
            else:
                self._outputs[aliased_tag] = args[x]
        self._io_checked = True

    def setup_mpi(self, comm=None):
        """
        Setup the MPI interface

        Parameters
        ----------
        comm: MPI communicator
            (default is None) An MPI comm object to use in preference to COMM_WORLD
        """
        mpi = self.config.get("mpi", False)

        if mpi:  # pragma: no cover
            try:
                # This isn't a ceci dependency, so give a sensible error message if not installed.
                import mpi4py.MPI
            except ImportError:
                print("ERROR: Using --mpi option requires mpi4py to be installed.")
                raise

        # For scripting and testing we allow an MPI communicator or anything
        # with the same API to be passed in directly, overriding the --mpi
        # flag.
        if comm is not None:
            self._parallel = MPI_PARALLEL
            self._comm = comm
            self._size = self._comm.Get_size()
            self._rank = self._comm.Get_rank()
        elif mpi:  # pragma: no cover
            self._parallel = MPI_PARALLEL
            self._comm = mpi4py.MPI.COMM_WORLD
            self._size = self._comm.Get_size()
            self._rank = self._comm.Get_rank()
        else:
            self._parallel = SERIAL
            self._comm = None
            self._size = 1
            self._rank = 0

        # If we are running under MPI but this subclass has enabled dask
        # then we note that here. It stops various MPI-specific things happening
        # later
        if (self._parallel == MPI_PARALLEL) and self.dask_parallel:
            self._parallel = DASK_PARALLEL

    pipeline_stages = {}
    incomplete_pipeline_stages = {}

    def __init_subclass__(cls, **kwargs):
        """
        Python 3.6+ provides a facility to automatically
        call a method (this one) whenever a new subclass
        is defined.  In this case we use that feature to keep
        track of all available pipeline stages, each of which is
        defined by a class.

        """
        super().__init_subclass__(**kwargs)

        # This is a hacky way of finding the file
        # where our stage was defined
        filename = sys.modules[cls.__module__].__file__

        stage_is_complete = (
            hasattr(cls, "inputs")
            and hasattr(cls, "outputs")
            and not getattr(cls.run, "__isabstractmethod__", False)
        )

        # If there isn't an explicit name already then set it here.
        # by default use the class name.
        if not hasattr(cls, "name"):  # pragma: no cover
            cls.name = cls.__name__
        if cls.name is None:  # pragma: no cover
            cls.name = cls.__name__

        if stage_is_complete:
            # Deal with duplicated class names
            if cls.name in cls.pipeline_stages and not cls.allow_reload:
                other = cls.pipeline_stages[cls.name][1]
                raise errors.DuplicateStageName(
                    "You created two pipeline stages with the"
                    f"name {cls.name}.\nOne was in {filename}\nand the "
                    f"other in {other}\nYou can either change the class "
                    "name or explicitly put a variable 'name' in the top"
                    "level of the class."
                )

            # Check for "config" in the inputs list - this is implicit
            for name, _ in cls.inputs:
                if name == "config":
                    raise errors.ReservedNameError(
                        "An input called 'config' is implicit in each pipeline "
                        "stage and should not be added explicitly.  Please update "
                        f"your pipeline stage called {cls.name} to remove/rename "
                        "the input called 'config'."
                    )

        # Check if user has over-written the config variable.
        # Quite a common error I make myself.
        if not isinstance(cls.config, property):
            raise errors.ReservedNameError(
                "You have a class variable called 'config', which "
                "is reserved in ceci for its own configuration. "
                "You may have meant to specify config_options?"
            )
        # Find the absolute path to the class defining the file
        path = pathlib.Path(filename).resolve()

        # Register the class
        if stage_is_complete:
            cls.pipeline_stages[cls.name] = (cls, path)
        else:
            cls.incomplete_pipeline_stages[cls.__name__] = (cls, path)

    #############################################
    # Life cycle-related methods and properties.
    #############################################

    @classmethod
    def get_stage(cls, name, module_name=None):
        """
        Return the PipelineStage subclass with the given name.

        This is used so that we do not need a new entry point __main__ function
        for each new stage - instead we can just use a single one which can query
        which class it should be using based on the name.

        If module_name is provided, this will import that module
        in order to load the required class.

        Returns
        -------
        cls: class
            The corresponding subclass
        """
        stage = cls.pipeline_stages.get(name)
        if stage is None:
            if module_name:
                __import__(module_name)
            stage = cls.pipeline_stages.get(name)

        # If not found, then check for incomplete stages
        if stage is None:
            if name in cls.incomplete_pipeline_stages:
                raise errors.IncompleteStage(
                    f"The stage {name} is not completely written. "
                    "Stages must specify 'inputs', 'outputs' as class variables "
                    f"and a 'run' method.\n{name} might be unfinished, or it might "
                    "be intended as a base for other classes and not to be run."
                )
            raise errors.StageNotFound(f"Unknown stage '{name}'")
        return stage[0]

    @classmethod
    def get_module(cls):
        """
        Return the path to the python package containing the current sub-class

        If we have a PipelineStage subclass defined in a module called "bar", in
        a package called "foo" e.g.:
        /path/to/foo/bar.py  <--   contains subclass "Baz"

        Then calling Baz.get_module() will return "foo.bar".

        We use this later to construct command lines like "python -m foo Baz"

        Returns
        -------
        module: str
            The module containing this class.
        """
        return cls.pipeline_stages[cls.name][0].__module__

    @classmethod
    def usage(cls):  # pragma: no cover
        """
        Print a usage message.
        """
        names = []
        docs = []
        for name, (stage, _) in cls.pipeline_stages.items():
            # find the first non-empty doc line, if there is one.
            try:
                doc_lines = [s.strip() for s in stage.__doc__.split("\n")]
                doc_lines = [d for d in doc_lines if d]
                doc = doc_lines[0]
            except (AttributeError, IndexError):
                doc = ""
            # cut off any very long lines
            if len(doc) > 100:
                doc = doc[:100] + " ..."
            # print the text
            names.append(name)
            docs.append(doc)

        # Make it look like a nice table by finding the maximum
        # length of the names, so that all the docs line up
        n = max(len(name) for name in names) + 1
        stage_texts = [f"- {name:{n}} - {d}" for name, d in zip(names, docs)]
        stage_text = "\n".join(stage_texts)

        try:
            module = cls.get_module().split(".")[0]
        except:  # pylint: disable=bare-except
            module = "<module_name>"
        sys.stderr.write(
            f"""
Usage: python -m {module} <stage_name> <stage_arguments>

If no stage_arguments are given then usage information
for the chosen stage will be given.

I currently know about these stages:

{stage_text}
"""
        )

    @classmethod
    def main(cls):
        """
        Create an instance of this stage and execute it with
        inputs and outputs taken from the command line
        """
        try:
            stage_name = sys.argv[1]
        except IndexError:  # pragma: no cover
            cls.usage()
            return 1
        if stage_name in ["--help", "-h"] and len(sys.argv) == 2:  # pragma: no cover
            cls.usage()
            return 1
        if stage_name.find(".") >= 0:
            tokens = stage_name.split(".")
            module_name = ".".join(tokens[:-1])
            stage_name = tokens[-1]
        else:
            module_name = None

        stage = cls.get_stage(stage_name, module_name)
        args = stage.parse_command_line()
        stage.execute(args)
        return 0

    @classmethod
    def parse_command_line(cls, cmd=None):
        """Set up and argument parser and parse the command line

        Parameters
        ----------
        cmd : str or None
            The command line to part (if None this will use the system arguments)

        Returns
        -------
        args : Namespace
            The resulting Mapping of arguement to values
        """
        import argparse

        parser = argparse.ArgumentParser(description=f"Run pipeline stage {cls.name}")
        parser.add_argument("stage_name")
        for conf, def_val in cls.config_options.items():
            if isinstance(def_val, StageParameter):
                opt_type = def_val.dtype
                def_val = def_val.default
            else:
                opt_type = def_val if isinstance(def_val, type) else type(def_val)
            if opt_type == bool:
                parser.add_argument(f"--{conf}", action="store_const", const=True)
                parser.add_argument(
                    f"--no-{conf}", dest=conf, action="store_const", const=False
                )
            elif opt_type == list:
                if not def_val:
                    out_type = str
                else:
                    out_type = (
                        def_val[0] if isinstance(def_val[0], type) else type(def_val[0])
                    )
                if out_type is str:  # pragma: no cover
                    parser.add_argument(
                        f"--{conf}", type=lambda string: string.split(",")
                    )
                elif out_type is int:  # pragma: no cover
                    parser.add_argument(
                        f"--{conf}",
                        type=lambda string: [int(i) for i in string.split(",")],
                    )
                elif out_type is float:
                    parser.add_argument(
                        f"--{conf}",
                        type=lambda string: [float(i) for i in string.split(",")],
                    )
                else:  # pragma: no cover
                    raise NotImplementedError(
                        "Only handles str, int and float list arguments"
                    )
            else:  # pragma: no cover
                parser.add_argument(f"--{conf}", type=opt_type)
        for inp in cls.input_tags():
            parser.add_argument(f"--{inp}")
        for out in cls.output_tags():
            parser.add_argument(f"--{out}")
        parser.add_argument(
            "--name",
            action="store",
            default=cls.name,
            type=str,
            help="Rename the stage",
        )
        parser.add_argument("--config")

        if cls.parallel:
            parser.add_argument(
                "--mpi", action="store_true", help="Set up MPI parallelism"
            )
        parser.add_argument(
            "--pdb", action="store_true", help="Run under the python debugger"
        )
        parser.add_argument(
            "--cprofile",
            action="store",
            default="",
            type=str,
            help="Profile the stage using the python cProfile tool",
        )
        parser.add_argument(
            "--memmon",
            type=int,
            default=0,
            help="Report memory use. Argument gives interval in seconds between reports",
        )

        # Error message we will return if --mpi used on a non-supported
        # stage.
        mpi_err = (
            "Error: you used the --mpi flag (or set MPI parallelism options) "
            f"for the stage {cls.name}, but that stage cannot be run in parallel."
        )

        if cmd is None:
            if ("--mpi" in sys.argv) and not cls.parallel:
                raise ValueError(mpi_err)
            ret_args = parser.parse_args()
        else:
            if ("--mpi" in cmd) and not cls.parallel:
                raise ValueError(mpi_err)
            ret_args = parser.parse_args(cmd)

        return ret_args

    @classmethod
    def execute(cls, args, comm=None):
        """
        Create an instance of this stage and run it
        with the specified inputs and outputs.

        This is calld by the main method.

        Parameters
        ----------
        args: namespace
            The argparse namespace for this subclass.
        """

        # Create the stage instance.  Running under dask this only
        # actually needs to happen for one process, but it's not a major
        # overhead and lets us do a whole bunch of other setup above
        stage = cls(args)
        stage.setup_mpi(comm)

        # This happens before dask is initialized
        start_time = datetime.datetime.now()
        if stage.rank == 0:
            start_time_text = start_time.isoformat(" ")
            print(f"Executing stage: {cls.name} @ {start_time_text}")

        if stage.is_dask():
            is_client = stage.start_dask()
            # worker and scheduler stages do not execute the
            # run method under dask
            if not is_client:
                return

        if args.cprofile:  # pragma: no cover
            profile = cProfile.Profile()
            profile.enable()

        if args.memmon:  # pragma: no cover
            monitor = MemoryMonitor.start_in_thread(interval=args.memmon)

        try:
            stage.run()
        except Exception as error:  # pragma: no cover
            if args.pdb:
                print(
                    "There was an exception - starting python debugger because you ran with --pdb"
                )
                print(error)
                pdb.post_mortem()
            else:
                if stage.rank == 0:
                    end_time = datetime.datetime.now()
                    end_time_text = end_time.isoformat(" ")
                    minutes = (end_time - start_time).total_seconds() / 60
                    print(
                        f"Stage failed: {cls.name} @ {end_time_text} after {minutes:.2f} minutes"
                    )
                raise
        finally:
            if args.memmon:  # pragma: no cover
                monitor.stop()
            if stage.is_dask():
                stage.stop_dask()

        # The default finalization renames any output files to their
        # final location, but subclasses can override to do other things too
        try:
            stage.finalize()
        except Exception as error:  # pragma: no cover
            if args.pdb:
                print(
                    "There was an exception in the finalization - starting python debugger because you ran with --pdb"
                )
                print(error)
                pdb.post_mortem()
            else:
                raise
        if args.cprofile:  # pragma: no cover
            profile.disable()
            profile.dump_stats(args.cprofile)
            profile.print_stats("cumtime")

        # Under dask the
        # the root process has gone off to become the scheduler,
        # and process 1 becomes the client which runs this code
        # and gets to this point
        if stage.rank == 0 or stage.is_dask():
            end_time = datetime.datetime.now()
            end_time_text = end_time.isoformat(" ")
            minutes = (end_time - start_time).total_seconds() / 60
            print(
                f"Stage complete: {cls.name} @ {end_time_text} took {minutes:.2f} minutes"
            )

    def finalize(self):
        """Finalize the stage, moving all its outputs to their final locations."""
        # Synchronize files so that everything is closed
        if self.is_mpi():  # pragma: no cover
            self.comm.Barrier()

        # Move files to their final path
        # Only the root process moves things, except under dask it is
        # process 1, which is the only process that reaches this point
        # (as noted above)
        if (self.rank == 0) or self.is_dask():
            for tag in self.output_tags():
                # find the old and new names
                self._finalize_tag(tag)

    def _finalize_tag(self, tag):
        """Finalize the data for a particular tag.

        This can be overridden by sub-classes for more complicated behavior
        """
        aliased_tag = self.get_aliased_tag(tag)
        temp_name = self.get_output(aliased_tag)
        final_name = self.get_output(aliased_tag, final_name=True)

        # it's not an error here if the path does not exist,
        # because that will be handled later.
        if pathlib.Path(temp_name).exists():
            # replace directories, rather than nesting more results
            if pathlib.Path(final_name).is_dir():  # pragma: no cover
                shutil.rmtree(final_name)
            shutil.move(temp_name, final_name)
        else:  # pragma: no cover
            sys.stderr.write(
                f"NOTE/WARNING: Expected output file {final_name} was not generated.\n"
            )
        return final_name

    #############################################
    # Parallelism-related methods and properties.
    #############################################
    @property
    def rank(self):
        """The rank of this process under MPI (0 if not running under MPI)"""
        return self._rank

    @property
    def size(self):
        """The number or processes under MPI (1 if not running under MPI)"""
        return self._size

    @property
    def comm(self):
        """The MPI communicator object (None if not running under MPI)"""
        return self._comm

    def is_parallel(self):
        """
        Returns True if the code is being run in parallel.
        Right now is_parallel() will return the same value as is_mpi(),
        but that may change in future if we implement other forms of
        parallelization.
        """
        return self._parallel != SERIAL

    def is_mpi(self):
        """
        Returns True if the stage is being run under MPI.
        """
        return self._parallel == MPI_PARALLEL

    def is_dask(self):
        """
        Returns True if the stage is being run in parallel with Dask.
        """
        return self._parallel == DASK_PARALLEL

    def start_dask(self):
        """
        Prepare dask to run under MPI. After calling this method
        only a single process, MPI rank 1 will continue to exeute code
        """

        # using the programmatic dask configuration system
        # does not seem to work. Presumably the loggers have already
        # been created by the time we modify the config. Doing it with
        # env vars seems to work. If the user has already set this then
        # we use that value. Otherwise we only want error logs
        key = "DASK_LOGGING__DISTRIBUTED"
        os.environ[key] = os.environ.get(key, "error")
        try:
            import dask
            import dask_mpi
            import dask.distributed
        except ImportError:  # pragma: no cover
            print(
                "ERROR: Using --mpi option on stages that use dask requires "
                "dask[distributed] and dask_mpi to be installed."
            )
            raise

        if self.size < 3:  # pragma: no cover
            raise ValueError(
                "Dask requires at least three processes. One becomes a scheduler "
                "process, one is a client that runs the code, and more are required "
                "as worker processes."
            )

        # This requires my fork until/unless they merge the PR, to allow
        # us to pass in these two arguments. In vanilla dask-mpi sys.exit
        # is called at the end of the event loop without returning to us.
        # After this point only a single process, MPI rank 1,
        # should continue to exeute code. The others enter an event
        # loop and return with is_client=False, which we return here
        # to tell the caller that they should not run everything.
        is_client = dask_mpi.initialize(comm=self.comm, exit=False)

        if is_client:
            # Connect this local process to remote workers.
            self.dask_client = dask.distributed.Client()
            # I don't yet know how to see this dashboard link at nersc
            print(f"Started dask. Diagnostics at {self.dask_client.dashboard_link}")

        return is_client

    @staticmethod
    def stop_dask():
        """
        End the dask event loop
        """
        from dask_mpi import send_close_signal

        send_close_signal()

    def split_tasks_by_rank(self, tasks):
        """Iterate through a list of items, yielding ones this process is responsible for/

        Tasks are allocated in a round-robin way.

        Parameters
        ----------
        tasks: iterable
            Tasks to split up

        """
        for i, task in enumerate(tasks):
            if i % self.size == self.rank:
                yield task

    def map_tasks_by_rank(self, function, inputs, allgather=False):
        """Run a function over a series of inputs, in parallel

        This mirrors the map function, and returns the equivalent of
        [function(input) for input in inputs], but executes in parallel.

        Parameters
        ----------
        function: Callable
            Function to be run on each item in inputs

        inputs: Iterable
            Any sequence of inputs, which should be the same
            on all processes. Or at least the same length:
            inputs not assigned to this process are ignored so
            you could get away with a dummy input for them.

        allgather: bool
            Whether to give all ranks the results (True) or just the
            root process (False). Default = False.

        Returns
        -------
        results: list
            A list of the results of calling the function on each input,
            in the same order as the input tasks
        """
        results = []
        # We keep track of the number of inputs manually rather
        # than calling len(inputs) because this allows inputs to
        # be an iterator.
        n = 0
        for i, inp in enumerate(inputs):
            n += 1
            if i % self.size == self.rank:
                results.append(function(inp))

        # If this is running in serial then the above just functions
        # like a basic map or list comprehension.
        if self.comm is not None:
            # Collate result as a list-of-lists, one sub-list for
            # each process
            if allgather:
                collected_results = self.comm.allgather(results)
            else:
                collected_results = self.comm.gather(results)
                if self.rank != 0:
                    return
            # convert the list-of-lists back into a single list
            # of results, returning to the original ordering.
            # The round-robin way we allocated them in the first
            # place is reversed by this.
            results = []
            for i in range(n):
                j = i % self.size
                k = i // self.size
                results.append(collected_results[j][k])

        return results

    def data_ranges_by_rank(self, n_rows, chunk_rows, parallel=True):
        """Split a number of rows by process.

        Given a total number of rows to read and a chunk size, yield
        the ranges within them that this process should handle.

        Parameters
        ----------
        n_rows: int
            Total number of rows to split up

        chunk_rows: int
            Size of each chunk to be read.

        Parallel: bool
            Whether to split data by rank or just give all procs all data.
            Default=True
        """
        n_chunks = n_rows // chunk_rows
        if n_chunks * chunk_rows < n_rows:  # pragma: no cover
            n_chunks += 1
        if parallel:
            it = self.split_tasks_by_rank(range(n_chunks))
        else:
            it = range(n_chunks)
        for i in it:
            start = i * chunk_rows
            end = min((i + 1) * chunk_rows, n_rows)
            yield start, end

    ##################################################
    # Input and output-related methods and properties.
    ##################################################

    def get_input(self, tag):
        """
        Return the path of an input file with the given tag,
        which can be aliased.
        """
        tag = self.get_aliased_tag(tag)
        return self._inputs[tag]



    def get_output(self, tag, final_name=False):
        """
        Return the path of an output file with the given tag,
        which can be aliased already.

        If final_name is False then use a temporary name - file will
        be moved to its final name at the end
        """

        tag = self.get_aliased_tag(tag)
        path = self._outputs[tag]

        # If not the final version, add a tag at the start of the filename
        if not final_name:
            p = pathlib.Path(path)
            p = p.parent / (IN_PROGRESS_PREFIX + p.name)
            path = str(p)
        return path

    def open_input(self, tag, wrapper=False, **kwargs):
        """
        Find and open an input file with the given tag, in read-only mode.

        For general files this will simply return a standard
        python file object.

        For specialized file types like FITS or HDF5 it will return
        a more specific object - see the types.py file for more info.

        """
        path = self.get_input(tag)
        input_class = self.get_input_type(tag)
        obj = input_class(path, "r", **kwargs)

        if wrapper:  # pragma: no cover
            return obj
        return obj.file

    def open_output(
        self, tag, wrapper=False, final_name=False, **kwargs
    ):  # pragma: no cover
        """
        Find and open an output file with the given tag, in write mode.

        If final_name is True then they will be opened using their final
        target output name.  Otherwise we will prepend "inprogress_" to their
        file name. This means we know that if the final file exists then it
        is completed.

        If wrapper is True this will return an instance of the class
        of the file as specified in the cls.outputs.  Otherwise it will
        return an open file object (standard python one or something more
        specialized).

        Parameters
        ----------

        tag: str
            Tag as listed in self.outputs

        wrapper: bool
            Default=False.  Whether to return a wrapped file

        final_name: bool
            Default=False. Whether to save to

        **kwargs:
            Extra args are passed on to the file's class constructor.

        """
        path = self.get_output(tag, final_name=final_name)
        output_class = self.get_output_type(tag)

        # HDF files can be opened for parallel writing
        # under MPI.  This checks if:
        # - we have been told to open in parallel
        # - we are actually running under MPI
        # and adds the flags required if all these are true
        run_parallel = kwargs.pop("parallel", False) and self.is_mpi()
        if run_parallel:
            kwargs["driver"] = "mpio"
            kwargs["comm"] = self.comm

            # XXX: This is also not a dependency, but it should be.
            #      Or even better would be to make it a dependency of descformats where it
            #      is actually used.
            import h5py

            if not h5py.get_config().mpi:
                print(
                    dedent(
                        """\
                Your h5py installation is not MPI-enabled.
                Options include:
                  1) Set nprocess to 1 for all stages
                  2) Upgrade h5py to use mpi.  See instructions here:
                     http://docs.h5py.org/en/latest/build.html#custom-installation
                Note: If using conda, the most straightforward way is to enable it is
                    conda install -c spectraldns h5py-parallel
                """
                    )
                )
                raise RuntimeError("h5py module is not MPI-enabled.")

        # Return an opened object representing the file
        obj = output_class(path, "w", **kwargs)
        if wrapper:
            return obj
        return obj.file

    @classmethod
    def inputs_(cls):
        """
        Return the dict of inputs
        """
        return cls.inputs  # pylint: disable=no-member

    @classmethod
    def outputs_(cls):
        """
        Return the dict of inputs
        """
        return cls.outputs  # pylint: disable=no-member

    @classmethod
    def output_tags(cls):
        """
        Return the list of output tags required by this stage
        """
        return [tag for tag, _ in cls.outputs_()]

    @classmethod
    def input_tags(cls):
        """
        Return the list of input tags required by this stage
        """
        return [tag for tag, _ in cls.inputs_()]

    def get_input_type(self, tag):
        """Return the file type class of an input file with the given tag."""
        tag = self.get_aliased_tag(tag)
        for t, dt in self.inputs_():
            t = self.get_aliased_tag(t)
            if t == tag:
                return dt
        raise ValueError(f"Tag {tag} is not a known input")  # pragma: no cover

    def get_output_type(self, tag):
        """Return the file type class of an output file with the given tag."""
        tag = self.get_aliased_tag(tag)
        for t, dt in self.outputs_():
            t = self.get_aliased_tag(t)
            if t == tag:
                return dt
        raise ValueError(f"Tag {tag} is not a known output")  # pragma: no cover

    ##################################################
    # Configuration-related methods and properties.
    ##################################################

    @property
    def instance_name(self):
        """Return the name associated to this particular instance of this stage"""
        return self._configs.get("name", self.name)

    @property
    def config(self):
        """
        Returns the configuration dictionary for this stage, aggregating command
        line options and optional configuration file.
        """
        return self._configs

    def read_config(self, args):
        """
        This function looks for the arguments of the pipeline stage using a
        combination of default values, command line options and separate
        configuration file.

        The order for resolving config options is first looking for a default
        value, then looking for a

        In case a mandatory argument (argument with no default) is missing,
        an exception is raised.

        Note that we recognize arguments with no default as the ones where
        self.config_options holds a type instead of a value.
        """
        # Try to load configuration file if provided
        import yaml

        config_file = self.get_input("config")

        # This is all the config information in the file, including
        # things for other stages
        if config_file is not None:
            with open(config_file) as _config_file:
                overall_config = yaml.safe_load(_config_file)
        else:
            overall_config = {}

        # The user can define global options that are inherited by
        # all the other sections if not already specified there.
        input_config = overall_config.get("global", {})

        # This is just the config info in the file for this stage.
        # It may be incomplete - there may be things specified on the
        # command line instead, or just using their default values
        stage_config = overall_config.get(self.instance_name, {})
        input_config.update(stage_config)

        self._configs.set_config(input_config, args)

    def get_config_dict(self, ignore=None, reduce_config=False):
        """Write the current configuration to a dict

        Parameters
        ----------
        ignore : dict or None
            Global parameters not to write
        reduce_config : bool
            If true, reduce the configuration by parsing out the inputs, outputs and global params

        Returns
        -------
        out_dict : dict
            The configuration
        """
        out_dict = {}
        if reduce_config:
            ignore_keys = self.input_tags() + self.output_tags() + ["config"]
        else:
            ignore_keys = []
        ignore = ignore or {}
        for key, val in self.config.items():
            if reduce_config:
                if key in ignore:
                    if ignore[key] == val:
                        continue
                if key in ignore_keys:
                    continue
            out_dict[key] = cast_to_streamable(val)
        return out_dict

    def find_inputs(self, pipeline_files):
        """Find and retrun all the inputs associated to this stage in the FileManager

        These are returned as a dictionary of tag : path pairs
        """
        ret_dict = {}
        for tag, _ in self.inputs_():
            aliased_tag = self.get_aliased_tag(tag)
            ret_dict[aliased_tag] = pipeline_files[aliased_tag]
        return ret_dict

    def find_outputs(self, outdir):
        """Find and retrun all the outputs associated to this stage

        These are returned as a dictionary of tag : path pairs
        """
        ret_dict = {}
        for tag, ftype in self.outputs_():
            aliased_tag = self.get_aliased_tag(tag)
            if not aliased_tag in self._outputs.keys():  # pragma: no cover
                self._outputs[aliased_tag] = ftype.make_name(aliased_tag)
            ret_dict[aliased_tag] = f"{outdir}/{self._outputs[aliased_tag]}"
        return ret_dict

    def print_io(self, stream=sys.stdout):
        """Print out the tags, paths and types for all the inputs and outputs of this stage"""
        stream.write("Inputs--------\n")
        for tag, ftype in self.inputs_():
            aliased_tag = self.get_aliased_tag(tag)
            stream.write(
                f"{tag:20} : {aliased_tag:20} :{str(ftype):20} : {self._inputs[tag]}\n"
            )
        stream.write("Outputs--------\n")
        for tag, ftype in self.outputs_():
            aliased_tag = self.get_aliased_tag(tag)
            stream.write(
                f"{tag:20} : {aliased_tag:20} :{str(ftype):20} : {self._outputs[aliased_tag]}\n"
            )

    def should_skip(self, run_config):
        """Return true if we should skip a stage b/c it's outputs already exist and we are in resume mode"""
        outputs = self.find_outputs(run_config["output_dir"]).values()
        already_run_stage = all(os.path.exists(output) for output in outputs)
        return already_run_stage and run_config["resume"]

    def already_finished(self):
        """Print a warning that a stage is being skipped"""
        print(f"Skipping stage {self.instance_name} because its outputs exist already")

    def iterate_fits(
        self, tag, hdunum, cols, chunk_rows, parallel=True
    ):  # pragma: no cover
        """
        Loop through chunks of the input data from a FITS file with the given tag

        TODO: add ceci tests of this functions
        Parameters
        ----------
        tag: str
            The tag from the inputs list to use

        hdunum: int
            The extension number to read

        cols: list
            The columns to read

        chunk_rows: int
            Number of columns to read and return at once

        parallel: bool
            Whether to split up data among processes (parallel=True) or give
            all processes all data (parallel=False).  Default = True.

        Returns
        -------
        it: iterator
            Iterator yielding (int, int, array) tuples of (start, end, data)
            data is a structured array.
        """
        fits = self.open_input(tag)
        ext = fits[hdunum]
        n = ext.get_nrows()
        for start, end in self.data_ranges_by_rank(n, chunk_rows, parallel=parallel):
            data = ext.read_columns(cols, rows=range(start, end))
            yield start, end, data

    def iterate_hdf(
        self, tag, group_name, cols, chunk_rows, parallel=True, longest=False
    ):
        """
        Loop through chunks of the input data from an HDF5 file with the given tag.

        All the selected columns must have the same length.

        Parameters
        ----------
        tag: str
            The tag from the inputs list to use

        group: str
            The group within the HDF5 file to use, looked up as
            file[group]

        cols: list
            The columns to read

        chunk_rows: int
            Number of columns to read and return at once

        parallel: bool
            Whether to split up data among processes (parallel=True) or give
            all processes all data (parallel=False).  Default = True.

        longest: bool
            Whether to allow mixed length arrays and keep going until the longest
            array is completed, returning empty arrays for shorter ones


        Returns
        -------
        it: iterator
            Iterator yielding (int, int, dict) tuples of (start, end, data)
        """
        import numpy as np

        hdf = self.open_input(tag)
        group = hdf[group_name]

        # Check all the columns are the same length
        N = [len(group[col]) for col in cols]
        n = max(N)
        if not longest:
            if not np.equal(N, n).all():
                raise ValueError(
                    f"Different columns among {cols} in file {tag} group {group_name}"
                    "are different sizes - if this is acceptable set longest=True"
                )

        # Iterate through the data providing chunks
        for start, end in self.data_ranges_by_rank(n, chunk_rows, parallel=parallel):
            data = {col: group[col][start:end] for col in cols}
            yield start, end, data

    ################################
    # Pipeline-related methods
    ################################

    @classmethod
    def generate_command(
        cls, inputs, config, outputs, aliases=None, instance_name=None
    ):
        """
        Generate a command line that will run the stage
        """
        module = cls.get_module()
        module = module.split(".")[0]

        if sys.modules[module].__file__:
            # Regular module, stage will be imported with module
            flags = [f"{cls.name}"]
        else:
            # Namescape module, use 'ceci' to the get main
            # and specify the full path
            flags = [f"{cls.get_module()}.{cls.name}"]
            module = "ceci"

        aliases = aliases or {}

        for tag, _ in cls.inputs_():
            aliased_tag = aliases.get(tag, tag)
            try:
                fpath = inputs[aliased_tag]
            except KeyError as msg:  # pragma: no cover
                raise ValueError(
                    f"Missing input location {aliased_tag} {str(inputs)}"
                ) from msg
            flags.append(f"--{tag}={fpath}")

        if instance_name is not None and instance_name != cls.name:
            flags.append(f"--name={instance_name}")

        flags.append(f"--config={config}")

        for tag, _ in cls.outputs_():
            aliased_tag = aliases.get(tag, tag)
            try:
                fpath = outputs[aliased_tag]
            except KeyError as msg:  # pragma: no cover
                raise ValueError(
                    f"Missing output location {aliased_tag} {str(outputs)}"
                ) from msg
            flags.append(f"--{tag}={fpath}")

        flags = "   ".join(flags)

        # We just return this, instead of wrapping it in a
        # parsl job
        cmd = f"python3 -m {module} {flags}"
        return cmd

    @classmethod
    def generate_cwl(cls, log_dir=None):
        """
        Produces a CWL App object which can then be exported to yaml
        """
        import cwl_utils.parser.cwl_v1_0 as cwlgen

        module = cls.get_module()
        module = module.split(".")[0]

        # Basic definition of the tool
        cwl_tool = cwlgen.CommandLineTool(
            [],
            [],
            id=cls.name,
            label=cls.name,
            baseCommand="python3",
            cwlVersion="v1.0",
            doc=cls.__doc__,
            arguments=[],
        )
        if log_dir is not None:
            cwl_tool.stdout = f"{cls.name}.out"
            cwl_tool.stderr = f"{cls.name}.err"

        # Adds the first input binding with the name of the module and pipeline stage
        input_arg = cwlgen.CommandLineBinding(position=-1, valueFrom=f"-m{module}")
        cwl_tool.arguments.append(input_arg)
        input_arg = cwlgen.CommandLineBinding(position=0, valueFrom=f"{cls.name}")
        cwl_tool.arguments.append(input_arg)

        type_dict = {int: "int", float: "float", str: "string", bool: "boolean"}
        # Adds the parameters of the tool
        for opt, def_val in cls.config_options.items():

            # Handles special case of lists:
            if isinstance(def_val, list):
                v = def_val[0]
                param_type = {
                    "type": "array",
                    "items": type_dict[v]
                    if isinstance(v, type)
                    else type_dict[type(v)],
                }
                default = def_val if not isinstance(v, type) else None
                input_binding = cwlgen.CommandLineBinding(
                    prefix=f"--{opt}=", itemSeparator=",", separate=False
                )
            else:
                param_type = (
                    type_dict[def_val]
                    if isinstance(def_val, type)
                    else type_dict[type(def_val)]
                )
                default = def_val if not isinstance(def_val, type) else None
                if param_type == "boolean":
                    input_binding = cwlgen.CommandLineBinding(prefix=f"--{opt}")
                else:  # pragma: no cover
                    input_binding = cwlgen.CommandLineBinding(
                        prefix=f"--{opt}=", separate=False
                    )

            input_param = cwlgen.CommandInputParameter(
                opt,
                label=opt,
                type=param_type,
                inputBinding=input_binding,
                default=default,
                doc="Some documentation about this parameter",
            )

            # We are bypassing the cwlgen builtin type check for the special case
            # of arrays until that gets added to the standard
            if isinstance(def_val, list):
                input_param.type = param_type

            cwl_tool.inputs.append(input_param)

        # Add the inputs of the tool
        for i, inp in enumerate(cls.input_tags()):
            input_binding = cwlgen.CommandLineBinding(prefix=f"--{inp}")
            input_param = cwlgen.CommandInputParameter(
                inp,
                label=inp,
                type="File",
                format=cls.inputs[i][1].format,  # pylint: disable=no-member
                inputBinding=input_binding,
                doc="Some documentation about the input",
            )
            cwl_tool.inputs.append(input_param)

        # Adds the overall configuration file
        input_binding = cwlgen.CommandLineBinding(prefix="--config")
        input_param = cwlgen.CommandInputParameter(
            "config",
            label="config",
            type="File",
            format="http://edamontology.org/format_3750",
            inputBinding=input_binding,
            doc="Configuration file",
        )
        cwl_tool.inputs.append(input_param)

        # Add the definition of the outputs
        for i, out in enumerate(cls.output_tags()):
            output_name = cls.outputs[i][1].make_name(out)  # pylint: disable=no-member
            output_binding = cwlgen.CommandOutputBinding(glob=output_name)
            output = cwlgen.CommandOutputParameter(
                out,
                label=out,
                type="File",
                outputBinding=output_binding,
                format=cls.outputs[i][1].format,  # pylint: disable=no-member
                doc="Some results produced by the pipeline element",
            )
            cwl_tool.outputs.append(output)

        if log_dir is not None:
            output = cwlgen.CommandOutputParameter(
                f"{cls.name}@stdout",
                label="stdout",
                type="stdout",
                doc="Pipeline elements standard output",
            )
            cwl_tool.outputs.append(output)
            error = cwlgen.CommandOutputParameter(
                f"{cls.name}@stderr",
                label="stderr",
                type="stderr",
                doc="Pipeline elements standard output",
            )
            cwl_tool.outputs.append(error)

        # Potentially add more metadata
        # This requires a schema however...
        # metadata = {'name': cls.name,
        #         'about': 'Some additional info',
        #         'publication': [{'id': 'one_doi'}, {'id': 'another_doi'}],
        #         'license': ['MIT']}
        # cwl_tool.metadata = cwlgen.Metadata(**metadata)

        return cwl_tool



import yaml
from ceci import FlowChartPipeline
import os

def test_flow_chart():
    config = yaml.safe_load(open("tests/test.yml"))
    launcher_config = {"interval": 0.5, "name": "mini"}

    pipeline = FlowChartPipeline(config["stages"], launcher_config)

    run_config = {
        "log_dir": config["log_dir"],
        "output_dir": config["output_dir"],
        "resume": False,
        "flow_chart": "ceci_test_flow_chart.png",
    }

    pipeline.initialize(config["inputs"], run_config, config["config"])
    status = pipeline.run()

    assert os.path.exists("ceci_test_flow_chart.png")
    os.remove('ceci_test_flow_chart.png')



"""Ceci n'est pas une pipeline"""

from .stage import PipelineStage
from .pipeline import (
    Pipeline,
    MiniPipeline,
    ParslPipeline,
    DryRunPipeline,
    FlowChartPipeline,
)
from .main import prepare_for_pipeline, run_pipeline
from pkg_resources import DistributionNotFound
from pkg_resources import get_distribution

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:  # pragma: no cover
    # package is not installed
    pass



from ceci.pipeline import StageExecutionConfig
from ceci.sites.local import LocalSite
from ceci.sites.nersc import NerscBatchSite
import os
import pytest


class MockSite:
    def __init__(self):
        self.config = {"image": "abc", "volume": "def"}


def test_defaults():
    sec = StageExecutionConfig({"name": "a", "site": MockSite()})
    assert sec.nodes == 1
    assert sec.nprocess == 1
    assert sec.threads_per_process == 1
    assert sec.mem_per_process == 2
    assert sec.image == "abc"
    assert sec.volume == "def"


def test_local():
    site = LocalSite({})
    sec = StageExecutionConfig({"name": "a", "site": site})
    cmd1 = "echo 1"

    # should not start with docker/shifter, since no image specified
    cmd = site.command(cmd1, sec)

    # don't want to test too specifically here, since it may change
    assert "docker" not in cmd
    assert "shifter" not in cmd
    assert "OMP_NUM_THREADS=1" in cmd
    assert cmd1 in cmd


def test_docker():
    site = LocalSite({})
    sec = StageExecutionConfig(
        {
            "name": "a",
            "site": site,
            "image": "username/potato",
            "volume": "a:b",
            "threads_per_process": 4,
            "nprocess": 2,
        }
    )
    cmd1 = "echo 1"

    # should not start with docker/shifter, since no image specified
    cmd = site.command(cmd1, sec)

    # don't want to test too specifically here, since it may change
    assert "docker run" in cmd
    assert "username/potato" in cmd
    assert "-v a:b" in cmd
    assert "mpirun -n 2" in cmd
    assert "shifter" not in cmd
    assert "OMP_NUM_THREADS=4" in cmd
    assert cmd1 in cmd


def _test_nersc(job_id):
    site = NerscBatchSite({})
    # fake that we're runnng a job to avoid complaints
    initial = os.environ.get("SLURM_JOB_ID")
    if job_id:
        os.environ["SLURM_JOB_ID"] = "fake_job_id"
    elif initial is not None:
        del os.environ["SLURM_JOB_ID"]

    try:

        sec = StageExecutionConfig(
            {
                "name": "a",
                "site": site,
                "image": "username/potato",
                "volume": "a:b",
                "threads_per_process": 4,
                "nprocess": 2,
                "nodes": 3,
            }
        )
        cmd1 = "echo 1"

        # should not start with docker/shifter, since no image specified
        cmd = site.command(cmd1, sec)

        # don't want to test too specifically here, since it may change
        assert "shifter" in cmd
        assert "--image username/potato" in cmd
        assert "-V a:b" in cmd
        assert "srun -u -n 2" in cmd
        assert "--env OMP_NUM_THREADS=4" in cmd
        assert "--nodes 3" in cmd
        assert "--mpi" in cmd
        assert cmd1 in cmd
    finally:
        if job_id:
            if initial is None:
                del os.environ["SLURM_JOB_ID"]
            else:
                os.environ["SLURM_JOB_ID"] = initial
        elif initial is not None:
            os.environ["SLURM_JOB_ID"] = initial


def test_works():
    _test_nersc(True)


def test_warning():
    with pytest.raises(ValueError):
        _test_nersc(False)



import sys
from ceci.utils import remove_last, extra_paths
import pytest
import os


def test_remove_item():
    l = list("abcdea")
    remove_last(l, "a")
    assert l == list("abcde")

    l = list("abcde")
    remove_last(l, "b")
    assert l == list("acde")

    with pytest.raises(ValueError):
        remove_last([1, 2, 3], 4)


class MyError(Exception):
    pass


def test_extra_paths():
    p = "xxx111yyy222"
    orig_path = sys.path[:]
    orig_env = os.environ.get("PYTHONPATH", "")

    # check path is put in
    with extra_paths(p):
        assert sys.path[0] == p
        assert p in os.environ["PYTHONPATH"]

    # check everything back to normal
    # after with statement
    assert p not in sys.path
    assert sys.path == orig_path
    assert p not in os.environ["PYTHONPATH"]
    assert os.environ["PYTHONPATH"] == orig_env

    # check that an exception does not interfere
    # with this
    try:
        with extra_paths(p):
            assert sys.path[0] == p
            raise MyError("x")
    except MyError:
        pass

    assert p not in sys.path
    assert sys.path == orig_path

    # now putting the item at the end not the start
    with extra_paths(p, start=False):
        assert sys.path[-1] == p

    assert p not in sys.path
    assert sys.path == orig_path

    try:
        with extra_paths(p, start=False):
            assert sys.path[-1] == p
            raise MyError("x")
    except MyError:
        pass

    assert p not in sys.path
    assert sys.path == orig_path

    # now agan with a list of paths
    p = ["xxx111yyy222", "aaa222333"]
    with extra_paths(p):
        assert sys.path[0] == p[1]
        assert sys.path[1] == p[0]

    for p1 in p:
        assert p1 not in sys.path
    assert sys.path == orig_path

    try:
        with extra_paths(p):
            assert sys.path[0] == p[1]
            assert sys.path[1] == p[0]
            raise MyError("x")
    except MyError:
        pass

    for p1 in p:
        assert p1 not in sys.path
    assert sys.path == orig_path

    # now agan with a list of paths, at the end
    p = ["xxx111yyy222", "aaa222333"]
    with extra_paths(p, start=False):
        assert sys.path[-1] == p[1]
        assert sys.path[-2] == p[0]

    for p1 in p:
        assert p1 not in sys.path
    assert sys.path == orig_path

    try:
        with extra_paths(p, start=False):
            assert sys.path[-1] == p[1]
            assert sys.path[-2] == p[0]
            raise MyError("x")
    except MyError:
        pass

    assert p not in sys.path
    assert sys.path == orig_path

    # check that if the user removes the path
    # themselves then it is okay
    p = ["xxx111yyy222", "aaa222333"]
    with extra_paths(p, start=True):
        sys.path.remove("xxx111yyy222")

    assert sys.path == orig_path

    # check only one copy is removed
    sys.path.append("aaa")
    tmp_paths = sys.path[:]
    p = "aaa"
    with extra_paths(p, start=True):
        pass

    assert sys.path == tmp_paths

    with extra_paths(p, start=False):
        pass



"""ceci specific error types"""


class DuplicateStageName(TypeError):
    """Raised when a stage name is reused in a pipeline"""


class IncompleteStage(TypeError):
    """Raised when a stage is missing required methods or attributes"""


class StageNotFound(ValueError):
    """Raise when a stage is not found in a pipeline"""


class ReservedNameError(TypeError):
    """Raised when a parameter is given a reserved name"""
